@@ -1,6 +1,6 @@
 /*-
  * Copyright (c) 2006 Verdens Gang AS
- * Copyright (c) 2006-2007 Linpro AS
+ * Copyright (c) 2006-2008 Linpro AS
  * All rights reserved.
  *
  * Author: Poul-Henning Kamp <phk@phk.freebsd.dk>
@@ -26,13 +26,12 @@
  * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
  * SUCH DAMAGE.
  *
- * $Id: cache_acceptor.c 2239 2007-11-08 21:30:05Z des $
+ * $Id: cache_acceptor.c 3598 2009-02-05 10:10:58Z tfheen $
  *
- * XXX: We need to pass sessions back into the event engine when they are
- * reused.  Not sure what the most efficient way is for that.  For now
- * write the session pointer to a pipe which the event engine monitors.
  */
 
+#include "config.h"
+
 #include <stdio.h>
 #include <errno.h>
 #include <poll.h>
@@ -44,32 +43,29 @@
 #include <sys/types.h>
 #include <sys/socket.h>
 
-#ifndef HAVE_SRANDOMDEV
-#include "compat/srandomdev.h"
-#endif
-
-#include "heritage.h"
+#include "cli.h"
+#include "cli_priv.h"
 #include "shmlog.h"
 #include "cache.h"
 #include "cache_acceptor.h"
 
-
-static struct acceptor *vca_acceptors[] = {
+static struct acceptor * const vca_acceptors[] = {
 #if defined(HAVE_KQUEUE)
 	&acceptor_kqueue,
 #endif
 #if defined(HAVE_EPOLL_CTL)
 	&acceptor_epoll,
 #endif
-#if defined(HAVE_POLL)
-	&acceptor_poll,
+#if defined(HAVE_PORT_CREATE)
+	&acceptor_ports,
 #endif
+	&acceptor_poll,
 	NULL,
 };
 
-static struct acceptor *vca_act;
+static struct acceptor const *vca_act;
 
-static pthread_t 	vca_thread_acct;
+static pthread_t	vca_thread_acct;
 static struct timeval	tv_sndtimeo;
 static struct timeval	tv_rcvtimeo;
 static struct linger	linger;
@@ -113,10 +109,24 @@ sock_test(int fd)
 void
 VCA_Prep(struct sess *sp)
 {
+	char addr[TCP_ADDRBUFSIZE];
+	char port[TCP_PORTBUFSIZE];
+
 
 	TCP_name(sp->sockaddr, sp->sockaddrlen,
-	    sp->addr, sizeof sp->addr, sp->port, sizeof sp->port);
-	VSL(SLT_SessionOpen, sp->fd, "%s %s", sp->addr, sp->port);
+	    addr, sizeof addr, port, sizeof port);
+	sp->addr = WS_Dup(sp->ws, addr);
+	sp->port = WS_Dup(sp->ws, port);
+	if (params->log_local_addr) {
+		AZ(getsockname(sp->fd, sp->mysockaddr, &sp->mysockaddrlen));
+		TCP_name(sp->mysockaddr, sp->mysockaddrlen,
+		    addr, sizeof addr, port, sizeof port);
+		VSL(SLT_SessionOpen, sp->fd, "%s %s %s %s",
+		    sp->addr, sp->port, addr, port);
+	} else {
+		VSL(SLT_SessionOpen, sp->fd, "%s %s %s",
+		    sp->addr, sp->port, sp->mylsock->name);
+	}
 	sp->acct.first = sp->t_open;
 	if (need_test)
 		sock_test(sp->fd);
@@ -144,10 +154,13 @@ vca_acct(void *arg)
 	socklen_t l;
 	struct sockaddr_storage addr_s;
 	struct sockaddr *addr;
-	int i, j;
+	int i;
 	struct pollfd *pfd;
 	struct listen_sock *ls;
+	unsigned u;
+	double now;
 
+	THR_SetName("cache-acceptor");
 	(void)arg;
 
 	/* Set up the poll argument */
@@ -155,6 +168,9 @@ vca_acct(void *arg)
 	AN(pfd);
 	i = 0;
 	VTAILQ_FOREACH(ls, &heritage.socks, list) {
+		if (ls->sock < 0)
+			continue;
+		AZ(listen(ls->sock, params->listen_depth));
 		AZ(setsockopt(ls->sock, SOL_SOCKET, SO_LINGER,
 		    &linger, sizeof linger));
 		pfd[i].events = POLLIN;
@@ -167,35 +183,58 @@ vca_acct(void *arg)
 		if (params->send_timeout != tv_sndtimeo.tv_sec) {
 			need_test = 1;
 			tv_sndtimeo.tv_sec = params->send_timeout;
-			VTAILQ_FOREACH(ls, &heritage.socks, list) 
+			VTAILQ_FOREACH(ls, &heritage.socks, list) {
+				if (ls->sock < 0)
+					continue;
 				AZ(setsockopt(ls->sock, SOL_SOCKET,
 				    SO_SNDTIMEO,
 				    &tv_sndtimeo, sizeof tv_sndtimeo));
+			}
 		}
 #endif
 #ifdef SO_RCVTIMEO_WORKS
 		if (params->sess_timeout != tv_rcvtimeo.tv_sec) {
 			need_test = 1;
 			tv_rcvtimeo.tv_sec = params->sess_timeout;
-			VTAILQ_FOREACH(ls, &heritage.socks, list) 
+			VTAILQ_FOREACH(ls, &heritage.socks, list) {
+				if (ls->sock < 0)
+					continue;
 				AZ(setsockopt(ls->sock, SOL_SOCKET,
 				    SO_RCVTIMEO,
 				    &tv_rcvtimeo, sizeof tv_rcvtimeo));
+			}
 		}
 #endif
 		i = poll(pfd, heritage.nsocks, 1000);
-		for (j = 0; j < heritage.nsocks; j++) {
-			if (pfd[j].revents == 0)
+		now = TIM_real();
+		u = 0;
+		VTAILQ_FOREACH(ls, &heritage.socks, list) {
+			if (ls->sock < 0)
+				continue;
+			if (pfd[u++].revents == 0)
 				continue;
 			VSL_stats->client_conn++;
 			l = sizeof addr_s;
 			addr = (void*)&addr_s;
-			i = accept(pfd[j].fd, addr, &l);
+			i = accept(ls->sock, addr, &l);
 			if (i < 0) {
-				if (errno != EAGAIN) {
-					VSL(SLT_Debug, pfd[j].fd,
-					    "Accept failed errno=%d", errno);
+				switch (errno) {
+				case EAGAIN:
+				case ECONNABORTED:
+					break;
+				case EMFILE:
+					VSL(SLT_Debug, ls->sock,
+					    "Too many open files "
+					    "when accept(2)ing. Sleeping.");
+					TIM_sleep(
+					    params->accept_fd_holdoff * 1000.0);
+					break;
+				default:
+					VSL(SLT_Debug, ls->sock,
+					    "Accept failed: %s",
+					    strerror(errno));
 					/* XXX: stats ? */
+					break;
 				}
 				continue;
 			}
@@ -204,9 +243,9 @@ vca_acct(void *arg)
 
 			sp->fd = i;
 			sp->id = i;
-			sp->t_open = TIM_real();
+			sp->t_open = now;
+			sp->mylsock = ls;
 
-			http_RecvPrep(sp->http);
 			sp->step = STP_FIRST;
 			WRK_QueueSession(sp);
 		}
@@ -216,34 +255,25 @@ vca_acct(void *arg)
 /*--------------------------------------------------------------------*/
 
 void
-vca_handover(struct sess *sp, int bad)
+vca_handover(struct sess *sp, int status)
 {
 
-	if (bad) {
-		vca_close_session(sp,
-		    bad == 1 ? "overflow" : "no request");
+	switch (status) {
+	case -2:
+		vca_close_session(sp, "blast");
 		SES_Delete(sp);
-		return;
-	}
-	sp->step = STP_RECV;
-	WRK_QueueSession(sp);
-}
-
-/*--------------------------------------------------------------------*/
-
-int
-vca_pollsession(struct sess *sp)
-{
-	int i;
-
-	i = http_RecvSome(sp->fd, sp->http);
-	if (i < 1)
-		return (i);
-	if (i == 1)
-		vca_close_session(sp, "overflow");
-	else if (i == 2)
+		break;
+	case -1:
 		vca_close_session(sp, "no request");
-	return (1);
+		SES_Delete(sp);
+		break;
+	case 1:
+		sp->step = STP_START;
+		WRK_QueueSession(sp);
+		break;
+	default:
+		INCOMPL();
+	}
 }
 
 /*--------------------------------------------------------------------*/
@@ -253,7 +283,7 @@ vca_close_session(struct sess *sp, const char *why)
 {
 	int i;
 
-	VSL(SLT_SessionClose, sp->fd, why);
+	VSL(SLT_SessionClose, sp->id, "%s", why);
 	if (sp->fd >= 0) {
 		i = close(sp->fd);
 		assert(i == 0 || errno != EBADF);	/* XXX EINVAL seen */
@@ -268,28 +298,82 @@ vca_return_session(struct sess *sp)
 	CHECK_OBJ_NOTNULL(sp, SESS_MAGIC);
 	AZ(sp->obj);
 	AZ(sp->vcl);
-	if (sp->fd < 0)
-		SES_Delete(sp);
+	assert(sp->fd >= 0);
+	/*
+	 * Set nonblocking in the worker-thread, before passing to the
+	 * acceptor thread, to reduce syscall density of the latter.
+	 */
+	TCP_nonblocking(sp->fd);
+	if (vca_act->pass == NULL)
+		assert(sizeof sp == write(vca_pipes[1], &sp, sizeof sp));
 	else
-		 assert(sizeof sp == write(vca_pipes[1], &sp, sizeof sp));
+		vca_act->pass(sp);
 }
 
 
 /*--------------------------------------------------------------------*/
 
+static void
+ccf_start(struct cli *cli, const char * const *av, void *priv)
+{
+
+	(void)cli;
+	(void)av;
+	(void)priv;
+
+	if (vca_act == NULL)
+		vca_act = vca_acceptors[0];
+
+	AN(vca_act);
+	AN(vca_act->name);
+
+	if (vca_act->pass == NULL)
+		AZ(pipe(vca_pipes));
+	vca_act->init();
+	AZ(pthread_create(&vca_thread_acct, NULL, vca_acct, NULL));
+	VSL(SLT_Debug, 0, "Acceptor is %s", vca_act->name);
+}
+
+static struct cli_proto vca_cmds[] = {
+	{ CLI_SERVER_START,	ccf_start },
+	{ NULL }
+};
+
 void
 VCA_Init(void)
 {
 
+	CLI_AddFuncs(MASTER_CLI, vca_cmds);
+}
 
-	/* XXX: Add selector mechanism at some point */
-	vca_act = vca_acceptors[0];
+void
+VCA_tweak_acceptor(struct cli *cli, const char *arg)
+{
+	int i;
 
-	if (vca_act->name == NULL) {
-		fprintf(stderr, "No acceptor in program\n");
-		exit (2);
+	if (arg == NULL) {
+		if (vca_act == NULL)
+			cli_out(cli, "default");
+		else
+			cli_out(cli, "%s", vca_act->name);
+
+		cli_out(cli, " (");
+		for (i = 0; vca_acceptors[i] != NULL; i++)
+			cli_out(cli, "%s%s", i == 0 ? "" : ", ",
+			    vca_acceptors[i]->name);
+		cli_out(cli, ")");
+		return;
+	}
+	if (!strcmp(arg, "default")) {
+		vca_act = NULL;
+		return;
+	}
+	for (i = 0; vca_acceptors[i]->name; i++) {
+		if (!strcmp(arg, vca_acceptors[i]->name)) {
+			vca_act = vca_acceptors[i];
+			return;
+		}
 	}
-	AZ(pipe(vca_pipes));
-	vca_act->init();
-	AZ(pthread_create(&vca_thread_acct, NULL, vca_acct, NULL));
+	cli_out(cli, "Unknown acceptor");
+	cli_result(cli, CLIS_PARAM);
 }
