@@ -1,6 +1,6 @@
 /*-
  * Copyright (c) 2006 Verdens Gang AS
- * Copyright (c) 2006-2007 Linpro AS
+ * Copyright (c) 2006-2008 Linpro AS
  * All rights reserved.
  *
  * Author: Poul-Henning Kamp <phk@phk.freebsd.dk>
@@ -26,16 +26,20 @@
  * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
  * SUCH DAMAGE.
  *
- * $Id: cache_acceptor.h 1693 2007-07-13 14:54:51Z des $
+ * $Id: cache_acceptor.h 3713 2009-02-09 13:25:00Z tfheen $
  */
 
 struct sess;
 
 typedef void acceptor_init_f(void);
+typedef void acceptor_pass_f(struct sess *);
+
+extern int vca_pipes[2];
 
 struct acceptor {
-	const char 		*name;
+	const char		*name;
 	acceptor_init_f		*init;
+	acceptor_pass_f		*pass;
 };
 
 #if defined(HAVE_EPOLL_CTL)
@@ -46,11 +50,11 @@ extern struct acceptor acceptor_epoll;
 extern struct acceptor acceptor_kqueue;
 #endif
 
-#if defined(HAVE_POLL)
 extern struct acceptor acceptor_poll;
+
+#if defined(HAVE_PORT_CREATE)
+extern struct acceptor acceptor_ports;
 #endif
 
 /* vca_acceptor.c */
 void vca_handover(struct sess *sp, int bad);
-int vca_pollsession(struct sess *sp);
-
