@@ -1,6 +1,6 @@
 /*-
  * Copyright (c) 2006 Verdens Gang AS
- * Copyright (c) 2006-2007 Linpro AS
+ * Copyright (c) 2006-2008 Linpro AS
  * All rights reserved.
  *
  * Author: Poul-Henning Kamp <phk@phk.freebsd.dk>
@@ -26,12 +26,13 @@
  * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
  * SUCH DAMAGE.
  *
- * $Id: cache_http.c 2239 2007-11-08 21:30:05Z des $
+ * $Id: cache_http.c 3732 2009-02-11 07:13:12Z tfheen $
  *
  * HTTP request storage and manipulation
  */
 
-#include <ctype.h>
+#include "config.h"
+
 #include <errno.h>
 #include <unistd.h>
 #include <stdarg.h>
@@ -40,8 +41,8 @@
 #include <string.h>
 #include <strings.h>
 
-#include "heritage.h"
 #include "shmlog.h"
+#include "vct.h"
 #include "cache.h"
 
 #ifndef HAVE_STRLCPY
@@ -52,48 +53,42 @@
 #include "http_headers.h"
 #undef HTTPH
 
-enum httptag {
-	HTTP_T_Request,
-	HTTP_T_Response,
-	HTTP_T_Status,
-	HTTP_T_URL,
-	HTTP_T_Protocol,
-	HTTP_T_Header,
-};
+/*lint -save -e773 not () */
+#define LOGMTX2(ax, bx, cx)	[bx] = SLT_##ax##cx
 
-#define LOGMTX2(ax, bx) 	\
-	[HTTP_T_##bx] = SLT_##ax##bx
-
-#define LOGMTX1(ax) { 		\
-	LOGMTX2(ax, Request),	\
-	LOGMTX2(ax, Response),	\
-	LOGMTX2(ax, Status),	\
-	LOGMTX2(ax, URL),	\
-	LOGMTX2(ax, Protocol),	\
-	LOGMTX2(ax, Header),	\
+#define LOGMTX1(ax) {					\
+	LOGMTX2(ax, HTTP_HDR_REQ,	Request),	\
+	LOGMTX2(ax, HTTP_HDR_RESPONSE,	Response),	\
+	LOGMTX2(ax, HTTP_HDR_STATUS,	Status),	\
+	LOGMTX2(ax, HTTP_HDR_URL,	URL),		\
+	LOGMTX2(ax, HTTP_HDR_PROTO,	Protocol),	\
+	LOGMTX2(ax, HTTP_HDR_FIRST,	Header),	\
 	}
 
-static enum shmlogtag logmtx[3][7] = {
+static const enum shmlogtag logmtx[][HTTP_HDR_FIRST + 1] = {
 	[HTTP_Rx] = LOGMTX1(Rx),
 	[HTTP_Tx] = LOGMTX1(Tx),
 	[HTTP_Obj] = LOGMTX1(Obj)
 };
+/*lint -restore */
 
 static enum shmlogtag
-http2shmlog(struct http *hp, enum httptag t)
+http2shmlog(const struct http *hp, int t)
 {
 
 	CHECK_OBJ_NOTNULL(hp, HTTP_MAGIC);
-	assert(/* hp->logtag >= HTTP_Rx && */hp->logtag <= HTTP_Obj);
-	assert(/* t >= HTTP_T_Request && */t <= HTTP_T_Header);
+	if (t > HTTP_HDR_FIRST)
+		t = HTTP_HDR_FIRST;
+	assert(hp->logtag >= HTTP_Rx && hp->logtag <= HTTP_Obj); /*lint !e685*/
+	assert(t >= HTTP_HDR_REQ && t <= HTTP_HDR_FIRST);
 	return (logmtx[hp->logtag][t]);
 }
 
 static void
-WSLH(struct worker *w, enum httptag t, unsigned fd, struct http *hp, int hdr)
+WSLH(struct worker *w, int fd, const struct http *hp, unsigned hdr)
 {
 
-	WSLR(w, http2shmlog(hp, t), fd, hp->hd[hdr].b, hp->hd[hdr].e);
+	WSLR(w, http2shmlog(hp, hdr), fd, hp->hd[hdr]);
 }
 
 /*--------------------------------------------------------------------*/
@@ -103,51 +98,13 @@ static struct http_msg {
 	unsigned	nbr;
 	const char	*txt;
 } http_msg[] = {
-	{ 101, "Switching Protocols" },
-	{ 200, "OK" },
-	{ 201, "Created" },
-	{ 202, "Accepted" },
-	{ 203, "Non-Authoritative Information" },
-	{ 204, "No Content" },
-	{ 205, "Reset Content" },
-	{ 206, "Partial Content" },
-	{ 300, "Multiple Choices" },
-	{ 301, "Moved Permanently" },
-	{ 302, "Found" },
-	{ 303, "See Other" },
-	{ 304, "Not Modified" },
-	{ 305, "Use Proxy" },
-	{ 306, "(Unused)" },
-	{ 307, "Temporary Redirect" },
-	{ 400, "Bad Request" },
-	{ 401, "Unauthorized" },
-	{ 402, "Payment Required" },
-	{ 403, "Forbidden" },
-	{ 404, "Not Found" },
-	{ 405, "Method Not Allowed" },
-	{ 406, "Not Acceptable" },
-	{ 407, "Proxy Authentication Required" },
-	{ 408, "Request Timeout" },
-	{ 409, "Conflict" },
-	{ 410, "Gone" },
-	{ 411, "Length Required" },
-	{ 412, "Precondition Failed" },
-	{ 413, "Request Entity Too Large" },
-	{ 414, "Request-URI Too Long" },
-	{ 415, "Unsupported Media Type" },
-	{ 416, "Requested Range Not Satisfiable" },
-	{ 417, "Expectation Failed" },
-	{ 500, "Internal Server Error" },
-	{ 501, "Not Implemented" },
-	{ 502, "Bad Gateway" },
-	{ 503, "Service Unavailable" },
-	{ 504, "Gateway Timeout" },
-	{ 505, "HTTP Version Not Supported" },
+#define HTTP_RESP(n, t)	{ n, t},
+#include "http_response.h"
 	{ 0, NULL }
 };
 
 const char *
-http_StatusMessage(int status)
+http_StatusMessage(unsigned status)
 {
 	struct http_msg *mp;
 
@@ -161,13 +118,12 @@ http_StatusMessage(int status)
 /*--------------------------------------------------------------------*/
 
 void
-http_Setup(struct http *hp, void *space, unsigned len)
+http_Setup(struct http *hp, struct ws *ws)
 {
 
-	assert(len > 0);
 	memset(hp, 0, sizeof *hp);
 	hp->magic = HTTP_MAGIC;
-	WS_Init(hp->ws, space, len);
+	hp->ws = ws;
 	hp->nhd = HTTP_HDR_FIRST;
 }
 
@@ -175,12 +131,11 @@ http_Setup(struct http *hp, void *space, unsigned len)
 
 
 static int
-http_IsHdr(struct http_hdr *hh, const char *hdr)
+http_IsHdr(const txt *hh, const char *hdr)
 {
 	unsigned l;
 
-	AN(hh->b);
-	AN(hh->e);
+	Tcheck(*hh);
 	AN(hdr);
 	l = hdr[0];
 	assert(l == strlen(hdr + 1));
@@ -192,13 +147,12 @@ http_IsHdr(struct http_hdr *hh, const char *hdr)
 /*--------------------------------------------------------------------*/
 
 static unsigned
-http_findhdr(struct http *hp, unsigned l, const char *hdr)
+http_findhdr(const struct http *hp, unsigned l, const char *hdr)
 {
 	unsigned u;
 
 	for (u = HTTP_HDR_FIRST; u < hp->nhd; u++) {
-		AN(hp->hd[u].b);
-		AN(hp->hd[u].e);
+		Tcheck(hp->hd[u]);
 		if (hp->hd[u].e < hp->hd[u].b + l + 1)
 			continue;
 		if (hp->hd[u].b[l] != ':')
@@ -211,101 +165,122 @@ http_findhdr(struct http *hp, unsigned l, const char *hdr)
 }
 
 int
-http_GetHdr(struct http *hp, const char *hdr, char **ptr)
+http_GetHdr(const struct http *hp, const char *hdr, char **ptr)
 {
 	unsigned u, l;
 	char *p;
 
 	l = hdr[0];
-	assert(l == strlen(hdr + 1));
+	diagnostic(l == strlen(hdr + 1));
 	assert(hdr[l] == ':');
 	hdr++;
 	u = http_findhdr(hp, l - 1, hdr);
 	if (u == 0) {
-		*ptr = NULL;
+		if (ptr != NULL)
+			*ptr = NULL;
 		return (0);
 	}
-	p = hp->hd[u].b + l;
-	while (isspace(*p))
-		p++;
-	*ptr = p;
+	if (ptr != NULL) {
+		p = hp->hd[u].b + l;
+		while (vct_issp(*p))
+			p++;
+		*ptr = p;
+	}
 	return (1);
 }
 
-/*--------------------------------------------------------------------*/
+/*--------------------------------------------------------------------
+ * Find a given headerfield, and if present and wanted, the beginning
+ * of its value.
+ */
 
 int
-http_GetHdrField(struct http *hp, const char *hdr, const char *field, char **ptr)
+http_GetHdrField(const struct http *hp, const char *hdr,
+    const char *field, char **ptr)
 {
-	char *h;
-	int fl;
+	char *h, *e;
+	unsigned fl;
 
+	if (ptr != NULL)
+		*ptr = NULL;
 	if (!http_GetHdr(hp, hdr, &h))
 		return (0);
+	AN(h);
+	e = strchr(h, '\0');
 	fl = strlen(field);
-	while (*h) {
-		if (isspace(*h)) {
+	while (h + fl <= e) {
+		/* Skip leading separators */
+		if (vct_issepctl(*h)) {
 			h++;
 			continue;
 		}
-		if (*h == ',') {
+		/* Check for substrings before memcmp() */
+		if ((h + fl == e || vct_issepctl(h[fl])) &&
+		    !memcmp(h, field, fl)) {
+			/* got it */
+			h += fl;
+			if (ptr != NULL) {
+				/* Skip whitespace, looking for '=' */
+				while (*h && vct_issp(*h))
+					h++;
+				if (*h == '=') {
+					h++;
+					while (*h && vct_issp(*h))
+						h++;
+					*ptr = h;
+				}
+			}
+			return (1);
+		}
+		/* Skip token */
+		while (*h && !vct_issepctl(*h))
 			h++;
-			continue;
-		}
-		if (memcmp(h, field, fl) ||
-		    isalpha(h[fl]) || h[fl] == '-') {
-			while (*h && !(isspace(*h) || *h == ','))
-				h++;
-			continue;
-		}
-		if (h[fl] == '=')
-			*ptr = &h[fl + 1];
-		else
-			*ptr = NULL;
-		return (1);
 	}
 	return (0);
 }
 
-/*--------------------------------------------------------------------*/
+/*--------------------------------------------------------------------
+ * XXX: redo with http_GetHdrField() ?
+ */
 
-void
-http_DoConnection(struct sess *sp)
+const char *
+http_DoConnection(struct http *hp)
 {
-	struct http *hp = sp->http;
 	char *p, *q;
-	int i;
+	const char *ret;
 	unsigned u;
 
 	if (!http_GetHdr(hp, H_Connection, &p)) {
-		if (strcmp(hp->hd[HTTP_HDR_PROTO].b, "HTTP/1.1"))
-			sp->doclose = "not HTTP/1.1";
-		return;
+		if (hp->protover < 1.1)
+			return ("not HTTP/1.1");
+		return (NULL);
 	}
+	ret = NULL;
 	for (; *p; p++) {
-		if (isspace(*p))
+		if (vct_issp(*p))
 			continue;
 		if (*p == ',')
 			continue;
 		for (q = p + 1; *q; q++)
-			if (*q == ',' || isspace(*q))
+			if (*q == ',' || vct_issp(*q))
 				break;
-		i = q - p;
-		if (i == 5 && !strncasecmp(p, "close", i))
-			sp->doclose = "Connection: close";
-		u = http_findhdr(hp, i, p);
+		u = pdiff(p, q);
+		if (u == 5 && !strncasecmp(p, "close", u))
+			ret = "Connection: close";
+		u = http_findhdr(hp, u, p);
 		if (u != 0)
 			hp->hdf[u] |= HDF_FILTER;
 		if (!*q)
 			break;
 		p = q;
 	}
+	return (ret);
 }
 
 /*--------------------------------------------------------------------*/
 
 int
-http_HdrIs(struct http *hp, const char *hdr, const char *val)
+http_HdrIs(const struct http *hp, const char *hdr, const char *val)
 {
 	char *p;
 
@@ -320,79 +295,18 @@ http_HdrIs(struct http *hp, const char *hdr, const char *val)
 /*--------------------------------------------------------------------*/
 
 int
-http_GetTail(struct http *hp, unsigned len, char **b, char **e)
+http_GetStatus(const struct http *hp)
 {
 
-	if (hp->pl_s >= hp->pl_e)
-		return (0);
-
-	if (len == 0)
-		len = hp->pl_e - hp->pl_s;
-
-	if (hp->pl_s + len > hp->pl_e)
-		len = hp->pl_e - hp->pl_s;
-	if (len == 0)
-		return (0);
-	*b = hp->pl_s;
-	*e = hp->pl_s + len;
-	hp->pl_s += len;
-	assert(hp->pl_s <= hp->pl_e);
-	return (1);
-}
-
-/*--------------------------------------------------------------------*/
-/* Read from fd, but soak up any tail first */
-
-int
-http_Read(struct http *hp, int fd, void *p, unsigned len)
-{
-	int i;
-	int u;
-	char *b = p;
-
-	u = 0;
-	if (hp->pl_s < hp->pl_e) {
-		u = hp->pl_e - hp->pl_s;
-		if (u > len)
-			u = len;
-		memcpy(b, hp->pl_s, u);
-		hp->pl_s += u;
-		b += u;
-		len -= u;
-	}
-	if (hp->pl_e == hp->pl_s)
-		hp->pl_s = hp->pl_e = NULL;
-	if (len > 0) {
-		i = read(fd, b, len);
-		if (i < 0)
-			return (i);
-		u += i;
-	}
-	return (u);
-}
-
-/*--------------------------------------------------------------------*/
-
-int
-http_GetStatus(struct http *hp)
-{
-
-	AN(hp->hd[HTTP_HDR_STATUS].b);
-	return (strtoul(hp->hd[HTTP_HDR_STATUS].b,
-	    NULL /* XXX */, 10));
+	Tcheck(hp->hd[HTTP_HDR_STATUS]);
+	return (hp->status);
 }
 
 const char *
-http_GetProto(struct http *hp)
+http_GetReq(const struct http *hp)
 {
-	AN(hp->hd[HTTP_HDR_PROTO].b);
-	return (hp->hd[HTTP_HDR_PROTO].b);
-}
 
-const char *
-http_GetReq(struct http *hp)
-{
-	AN(hp->hd[HTTP_HDR_REQ].b);
+	Tcheck(hp->hd[HTTP_HDR_REQ]);
 	return (hp->hd[HTTP_HDR_REQ].b);
 }
 
@@ -402,7 +316,7 @@ http_GetReq(struct http *hp)
  */
 
 static int
-http_dissect_hdrs(struct worker *w, struct http *hp, int fd, char *p)
+http_dissect_hdrs(struct worker *w, struct http *hp, int fd, char *p, txt t)
 {
 	char *q, *r;
 
@@ -412,15 +326,13 @@ http_dissect_hdrs(struct worker *w, struct http *hp, int fd, char *p)
 	hp->nhd = HTTP_HDR_FIRST;
 	hp->conds = 0;
 	r = NULL;		/* For FlexeLint */
-	assert(p < hp->rx_e);	/* http_header_complete() guarantees this */
-	for (; p < hp->rx_e; p = r) {
+	for (; p < t.e; p = r) {
 		/* XXX: handle continuation lines */
 		q = strchr(p, '\n');
 		assert(q != NULL);
 		r = q + 1;
 		if (q > p && q[-1] == '\r')
 			q--;
-		*q = '\0';
 		if (p == q)
 			break;
 
@@ -429,265 +341,173 @@ http_dissect_hdrs(struct worker *w, struct http *hp, int fd, char *p)
 		    p[2] == '-')
 			hp->conds = 1;
 
+		while (q > p && vct_issp(q[-1]))
+			q--;
+		*q = '\0';
+
 		if (hp->nhd < HTTP_HDR_MAX) {
 			hp->hdf[hp->nhd] = 0;
 			hp->hd[hp->nhd].b = p;
 			hp->hd[hp->nhd].e = q;
-			WSLH(w, HTTP_T_Header, fd, hp, hp->nhd);
+			WSLH(w, fd, hp, hp->nhd);
 			hp->nhd++;
 		} else {
 			VSL_stats->losthdr++;
-			WSLR(w, SLT_LostHeader, fd, p, q);
+			WSL(w, SLT_LostHeader, fd, "%.*s", q - p, p);
+			return (400);
 		}
 	}
 	return (0);
 }
 
-/*--------------------------------------------------------------------*/
-
-int
-http_DissectRequest(struct worker *w, struct http *hp, int fd)
-{
-	char *p;
-
-	CHECK_OBJ_NOTNULL(hp, HTTP_MAGIC);
-	/* Assert a NUL at rx_e */
-	assert(hp->rx_s < hp->rx_e);
-	hp->logtag = HTTP_Rx;
-
-	for (p = hp->rx_s ; isspace(*p); p++)
-		continue;
-
-	/* First, the request type (GET/HEAD etc) */
-	hp->hd[HTTP_HDR_REQ].b = p;
-	for (; isalpha(*p); p++)
-		;
-	hp->hd[HTTP_HDR_REQ].e = p;
-	WSLH(w, HTTP_T_Request, fd, hp, HTTP_HDR_REQ);
-	*p++ = '\0';
-
-	/* Next find the URI */
-	while (isspace(*p) && *p != '\n')
-		p++;
-	if (*p == '\n') {
-		WSLR(w, SLT_HttpGarbage, fd, hp->rx_s, hp->rx_e);
-		return (400);
-	}
-	hp->hd[HTTP_HDR_URL].b = p;
-	while (!isspace(*p))
-		p++;
-	hp->hd[HTTP_HDR_URL].e = p;
-	WSLH(w, HTTP_T_URL, fd, hp, HTTP_HDR_URL);
-	if (*p == '\n') {
-		WSLR(w, SLT_HttpGarbage, fd, hp->rx_s, hp->rx_e);
-		return (400);
-	}
-	*p++ = '\0';
-
-	/* Finally, look for protocol */
-	while (isspace(*p) && *p != '\n')
-		p++;
-	if (*p == '\n') {
-		WSLR(w, SLT_HttpGarbage, fd, hp->rx_s, hp->rx_e);
-		return (400);
-	}
-	hp->hd[HTTP_HDR_PROTO].b = p;
-	while (!isspace(*p))
-		p++;
-	hp->hd[HTTP_HDR_PROTO].e = p;
-	WSLH(w, HTTP_T_Protocol, fd, hp, HTTP_HDR_PROTO);
-	if (*p != '\n')
-		*p++ = '\0';
-	while (isspace(*p) && *p != '\n')
-		p++;
-	if (*p != '\n') {
-		WSLR(w, SLT_HttpGarbage, fd, hp->rx_s, hp->rx_e);
-		return (400);
-	}
-	*p++ = '\0';
-
-	return (http_dissect_hdrs(w, hp, fd, p));
-}
-
-/*--------------------------------------------------------------------*/
-
-int
-http_DissectResponse(struct worker *w, struct http *hp, int fd)
-{
-	char *p, *q;
-
-	CHECK_OBJ_NOTNULL(hp, HTTP_MAGIC);
-	/* Assert a NUL at rx_e */
-	assert(hp->rx_s < hp->rx_e);
-	hp->logtag = HTTP_Rx;
-
-	for (p = hp->rx_s ; isspace(*p); p++)
-		continue;
-
-	if (memcmp(p, "HTTP/1.", 7)) {
-		WSLR(w, SLT_HttpGarbage, fd, hp->rx_s, hp->rx_e);
-		return (400);
-	}
-	/* First, protocol */
-	hp->hd[HTTP_HDR_PROTO].b = p;
-	while (!isspace(*p))
-		p++;
-	hp->hd[HTTP_HDR_PROTO].e = p;
-	WSLH(w, HTTP_T_Protocol, fd, hp, HTTP_HDR_PROTO);
-	*p++ = '\0';
-
-	/* Next find the status */
-	while (isspace(*p))
-		p++;
-	hp->hd[HTTP_HDR_STATUS].b = p;
-	while (!isspace(*p))
-		p++;
-	hp->hd[HTTP_HDR_STATUS].e = p;
-	WSLH(w, HTTP_T_Status, fd, hp, HTTP_HDR_STATUS);
-	*p++ = '\0';
-
-	/* Next find the response */
-	while (isspace(*p))
-		p++;
-	hp->hd[HTTP_HDR_RESPONSE].b = p;
-	while (*p != '\n')
-		p++;
-	for (q = p; q > hp->hd[HTTP_HDR_RESPONSE].b &&
-	    isspace(q[-1]); q--)
-		continue;
-	*q = '\0';
-	hp->hd[HTTP_HDR_RESPONSE].e = q;
-	WSLH(w, HTTP_T_Response, fd, hp, HTTP_HDR_RESPONSE);
-	p++;
-
-	return (http_dissect_hdrs(w, hp, fd, p));
-}
-
 /*--------------------------------------------------------------------
- * Return nonzero if we have a complete HTTP request.
+ * Deal with first line of HTTP protocol message.
  */
 
 static int
-http_header_complete(struct http *hp)
+http_splitline(struct worker *w, int fd, struct http *hp,
+    const struct http_conn *htc, int h1, int h2, int h3)
 {
 	char *p;
 
+	CHECK_OBJ_NOTNULL(htc, HTTP_CONN_MAGIC);
 	CHECK_OBJ_NOTNULL(hp, HTTP_MAGIC);
-	assert(*hp->rx_e == '\0');
-	/* Skip any leading white space */
-	for (p = hp->rx_s ; p < hp->rx_e && isspace(*p); p++)
+
+	/* XXX: Assert a NUL at rx.e ? */
+	Tcheck(htc->rxbuf);
+
+	/* Skip leading LWS */
+	for (p = htc->rxbuf.b ; vct_islws(*p); p++)
 		continue;
-	if (p >= hp->rx_e) {
-		hp->rx_e = hp->rx_s;
-		return (0);
+
+	/* First field cannot contain SP, CRLF or CTL */
+	hp->hd[h1].b = p;
+	for (; !vct_issp(*p); p++)
+		if (vct_isctl(*p))
+			return (400);
+	hp->hd[h1].e = p;
+
+	/* Skip SP */
+	for (; vct_issp(*p); p++)
+		;
+
+	/* Second field cannot contain LWS */
+	hp->hd[h2].b = p;
+	for (; !vct_islws(*p); p++)
+		;
+	hp->hd[h2].e = p;
+
+	if (!Tlen(hp->hd[h2]))
+		return (400);
+
+	/* Skip SP */
+	for (; vct_issp(*p); p++)
+		;
+
+	/* Third field is optional and cannot contain CTL */
+	if (!vct_iscrlf(*p)) {
+		hp->hd[h3].b = p;
+		for (; !vct_iscrlf(*p); p++)
+			if (vct_isctl(*p))
+				return (400);
+		hp->hd[h3].e = p;
+	} else {
+		hp->hd[h3].b = p;
+		hp->hd[h3].e = p;
 	}
-	while (1) {
-		/* XXX: we could save location of all linebreaks for later */
-		p = strchr(p, '\n');
-		if (p == NULL)
-			return (0);	/* XXX: Could cache p */
-		p++;
-		if (*p == '\r')
-			p++;
-		if (*p == '\n')
-			break;
+
+	/* Skip CRLF */
+	p += vct_skipcrlf(p);
+
+	*hp->hd[h1].e = '\0';
+	WSLH(w, fd, hp, h1);
+
+	*hp->hd[h2].e = '\0';
+	WSLH(w, fd, hp, h2);
+
+	if (hp->hd[h3].e != NULL) {
+		*hp->hd[h3].e = '\0';
+		WSLH(w, fd, hp, h3);
 	}
-	p++;
-	WS_ReleaseP(hp->ws, hp->rx_e);
-	if (p != hp->rx_e) {
-		hp->pl_s = p;
-		hp->pl_e = hp->rx_e;
-		hp->rx_e = p;
-	}
-	/* XXX: Check this stuff... */
-	return (1);
+
+	return (http_dissect_hdrs(w, hp, fd, p, htc->rxbuf));
 }
 
 /*--------------------------------------------------------------------*/
 
-void
-http_RecvPrep(struct http *hp)
+static void
+http_ProtoVer(struct http *hp)
 {
-	unsigned l;
 
-	CHECK_OBJ_NOTNULL(hp, HTTP_MAGIC);
-	WS_Assert(hp->ws);
-	WS_Reset(hp->ws);
-	WS_Reserve(hp->ws, 0);
-	hp->rx_s = hp->ws->f;
-	hp->rx_e = hp->rx_s;
-	if (hp->pl_s != NULL) {
-		assert(hp->pl_s < hp->pl_e);
-		l = hp->pl_e - hp->pl_s;
-		memmove(hp->rx_s, hp->pl_s, l);
-		hp->rx_e = hp->rx_s + l;
-		hp->pl_s = hp->pl_e = NULL;
-	}
-	*hp->rx_e = '\0';
-}
-
-int
-http_RecvPrepAgain(struct http *hp)
-{
-	http_RecvPrep(hp);
-	if (hp->rx_s == hp->rx_e)
-		return (0);
-	return (http_header_complete(hp));
-}
-
-/*--------------------------------------------------------------------*/
-
-int
-http_RecvSome(int fd, struct http *hp)
-{
-	unsigned l;
-	int i;
-
-	l = (hp->ws->e - hp->rx_e) - 1;
-	l /= 2;		/* Don't fill all of workspace with read-ahead */
-	if (l <= 1) {
-		VSL(SLT_HttpError, fd, "Received too much");
-		VSLR(SLT_HttpGarbage, fd, hp->rx_s, hp->rx_e);
-		hp->rx_s = hp->rx_e = NULL;
-		WS_Release(hp->ws, 0);
-		return (1);
-	}
-	errno = 0;
-	i = read(fd, hp->rx_e, l - 1);
-	if (i > 0) {
-		hp->rx_e += i;
-		*hp->rx_e = '\0';
-		if (http_header_complete(hp))
-			return(0);
-		return (-1);
-	}
-
-	if (hp->rx_e != hp->rx_s) {
-		VSL(SLT_HttpError, fd,
-		    "Received (only) %d bytes, errno %d",
-		    hp->rx_e - hp->rx_s, errno);
-		VSLR(SLT_Debug, fd, hp->rx_s, hp->rx_e);
-	} else if (errno == 0)
-		VSL(SLT_HttpError, fd, "Received nothing");
+	if (!strcmp(hp->hd[HTTP_HDR_PROTO].b, "HTTP/1.0"))
+		hp->protover = 1.0;
+	else if (!strcmp(hp->hd[HTTP_HDR_PROTO].b, "HTTP/1.1"))
+		hp->protover = 1.1;
 	else
-		VSL(SLT_HttpError, fd, "Received errno %d", errno);
-	hp->rx_s = hp->rx_e = NULL;
-	WS_Release(hp->ws, 0);
-	return(2);
+		hp->protover = 0.9;
+}
+
+
+/*--------------------------------------------------------------------*/
+
+int
+http_DissectRequest(struct sess *sp)
+{
+	struct http_conn *htc;
+	struct http *hp;
+	int i;
+
+	CHECK_OBJ_NOTNULL(sp, SESS_MAGIC);
+	htc = sp->htc;
+	CHECK_OBJ_NOTNULL(htc, HTTP_CONN_MAGIC);
+	hp = sp->http;
+	CHECK_OBJ_NOTNULL(hp, HTTP_MAGIC);
+
+	hp->logtag = HTTP_Rx;
+
+	i = http_splitline(sp->wrk, sp->fd, hp, htc,
+	    HTTP_HDR_REQ, HTTP_HDR_URL, HTTP_HDR_PROTO);
+	if (i != 0) {
+		WSPR(sp, SLT_HttpGarbage, htc->rxbuf);
+		return (i);
+	}
+	http_ProtoVer(hp);
+	return (i);
 }
 
 /*--------------------------------------------------------------------*/
 
 int
-http_RecvHead(struct http *hp, int fd)
+http_DissectResponse(struct worker *w, const struct http_conn *htc,
+    struct http *hp)
 {
 	int i;
 
+	CHECK_OBJ_NOTNULL(htc, HTTP_CONN_MAGIC);
 	CHECK_OBJ_NOTNULL(hp, HTTP_MAGIC);
-	http_RecvPrep(hp);
-	do
-		i = http_RecvSome(fd, hp);
-	while (i == -1);
+	hp->logtag = HTTP_Rx;
+
+	i = http_splitline(w, htc->fd, hp, htc,
+	    HTTP_HDR_PROTO, HTTP_HDR_STATUS, HTTP_HDR_RESPONSE);
+
+	if (i != 0 || memcmp(hp->hd[HTTP_HDR_PROTO].b, "HTTP/1.", 7))
+		WSLR(w, SLT_HttpGarbage, htc->fd, htc->rxbuf);
+	if (i != 0) {
+		if (hp->status == 0)
+			hp->status = i;
+	} else {
+		hp->status =
+		    strtoul(hp->hd[HTTP_HDR_STATUS].b, NULL /* XXX */, 10);
+	}
+	http_ProtoVer(hp);
+	if (hp->hd[HTTP_HDR_RESPONSE].b == NULL ||
+	    !Tlen(hp->hd[HTTP_HDR_RESPONSE])) {
+		/* Backend didn't send a response string, use the standard */
+		hp->hd[HTTP_HDR_RESPONSE].b =
+		    TRUST_ME(http_StatusMessage(hp->status));
+		hp->hd[HTTP_HDR_RESPONSE].e =
+		    strchr(hp->hd[HTTP_HDR_RESPONSE].b, '\0');
+	}
 	return (i);
 }
 
@@ -699,41 +519,47 @@ http_SetH(struct http *to, unsigned n, const char *fm)
 
 	assert(n < HTTP_HDR_MAX);
 	AN(fm);
-	to->hd[n].b = (void*)(uintptr_t)fm;
-	to->hd[n].e = (void*)(uintptr_t)strchr(fm, '\0');
+	to->hd[n].b = TRUST_ME(fm);
+	to->hd[n].e = strchr(to->hd[n].b, '\0');
 	to->hdf[n] = 0;
 }
 
 static void
-http_copyh(struct http *to, struct http *fm, unsigned n)
+http_copyh(struct http *to, const struct http *fm, unsigned n)
 {
 
 	assert(n < HTTP_HDR_MAX);
-	AN(fm->hd[n].b);
-	to->hd[n].b = fm->hd[n].b;
-	to->hd[n].e = fm->hd[n].e;
+	Tcheck(fm->hd[n]);
+	to->hd[n] = fm->hd[n];
 	to->hdf[n] = fm->hdf[n];
 }
 
 static void
-http_copyreq(struct http *to, struct http *fm, int transparent)
+http_copyreq(struct http *to, const struct http *fm, int how)
 {
 
 	CHECK_OBJ_NOTNULL(fm, HTTP_MAGIC);
 	CHECK_OBJ_NOTNULL(to, HTTP_MAGIC);
-	if (transparent)
+
+	if ((how == HTTPH_R_PIPE) || (how == HTTPH_R_PASS)) {
 		http_copyh(to, fm, HTTP_HDR_REQ);
-	else
-		http_SetH(to, HTTP_HDR_REQ, "GET");
-	http_copyh(to, fm, HTTP_HDR_URL);
-	if (transparent)
 		http_copyh(to, fm, HTTP_HDR_PROTO);
-	else
+	} else {
+		http_SetH(to, HTTP_HDR_REQ, "GET");
 		http_SetH(to, HTTP_HDR_PROTO, "HTTP/1.1");
+	}
+	http_copyh(to, fm, HTTP_HDR_URL);
 }
 
 void
-http_CopyResp(struct http *to, struct http *fm)
+http_ForceGet(struct http *to)
+{
+	if (strcmp(http_GetReq(to), "GET"))
+		http_SetH(to, HTTP_HDR_REQ, "GET");
+}
+
+void
+http_CopyResp(struct http *to, const struct http *fm)
 {
 
 	CHECK_OBJ_NOTNULL(fm, HTTP_MAGIC);
@@ -747,7 +573,8 @@ http_CopyResp(struct http *to, struct http *fm)
 }
 
 void
-http_SetResp(struct http *to, const char *proto, const char *status, const char *response)
+http_SetResp(struct http *to, const char *proto, const char *status,
+    const char *response)
 {
 
 	CHECK_OBJ_NOTNULL(to, HTTP_MAGIC);
@@ -757,33 +584,36 @@ http_SetResp(struct http *to, const char *proto, const char *status, const char
 }
 
 static void
-http_copyheader(struct worker *w, int fd, struct http *to, struct http *fm, unsigned n)
+http_copyheader(struct worker *w, int fd, struct http *to,
+    const struct http *fm, unsigned n)
 {
 
 	CHECK_OBJ_NOTNULL(fm, HTTP_MAGIC);
 	CHECK_OBJ_NOTNULL(to, HTTP_MAGIC);
 	assert(n < HTTP_HDR_MAX);
-	AN(fm->hd[n].b);
+	Tcheck(fm->hd[n]);
 	if (to->nhd < HTTP_HDR_MAX) {
-		to->hd[to->nhd].b = fm->hd[n].b;
-		to->hd[to->nhd].e = fm->hd[n].e;
+		to->hd[to->nhd] = fm->hd[n];
+		to->hdf[to->nhd] = 0;
 		to->nhd++;
 	} else  {
 		VSL_stats->losthdr++;
-		WSLH(w, SLT_LostHeader, fd, fm, n);
+		WSLR(w, SLT_LostHeader, fd, fm->hd[n]);
 	}
 }
 
 /*--------------------------------------------------------------------*/
 
 void
-http_FilterFields(struct worker *w, int fd, struct http *to, struct http *fm, unsigned how)
+http_FilterFields(struct worker *w, int fd, struct http *to,
+    const struct http *fm, unsigned how)
 {
 	unsigned u;
 
 	CHECK_OBJ_NOTNULL(fm, HTTP_MAGIC);
 	CHECK_OBJ_NOTNULL(to, HTTP_MAGIC);
 	to->nhd = HTTP_HDR_FIRST;
+	to->status = fm->status;
 	for (u = HTTP_HDR_FIRST; u < fm->nhd; u++) {
 		if (fm->hdf[u] & HDF_FILTER)
 			continue;
@@ -803,23 +633,19 @@ http_FilterHeader(struct sess *sp, unsigned how)
 {
 	struct bereq *bereq;
 	struct http *hp;
-	char *b;
 
-        bereq = vbe_new_bereq();
-        AN(bereq);
-        hp = bereq->http;
-        hp->logtag = HTTP_Tx;
+	AZ(sp->bereq);
+	bereq = VBE_new_bereq();
+	AN(bereq);
+	hp = bereq->http;
+	hp->logtag = HTTP_Tx;
 
-	http_copyreq(hp, sp->http,
-	    (how == HTTPH_R_PIPE) || (how == HTTPH_R_PASS));
+	http_copyreq(hp, sp->http, how);
 	http_FilterFields(sp->wrk, sp->fd, hp, sp->http, how);
 	http_PrintfHeader(sp->wrk, sp->fd, hp, "X-Varnish: %u", sp->xid);
 	http_PrintfHeader(sp->wrk, sp->fd, hp,
-	    "X-Forwarded-for: %s", sp->addr);
-	if (!http_GetHdr(hp, H_Host, &b)) {
-		http_PrintfHeader(sp->wrk, sp->fd, hp, "Host: %s",
-		    sp->backend->hostname);
-	}
+	    "X-Forwarded-For: %s", sp->addr);
+
 	sp->bereq = bereq;
 }
 
@@ -832,39 +658,24 @@ void
 http_CopyHome(struct worker *w, int fd, struct http *hp)
 {
 	unsigned u, l;
-	enum httptag htt;
 	char *p;
 
 	for (u = 0; u < hp->nhd; u++) {
 		if (hp->hd[u].b == NULL)
 			continue;
-		switch (u) {
-		case HTTP_HDR_PROTO:
-			htt = HTTP_T_Protocol;
-			break;
-		case HTTP_HDR_STATUS:
-			htt = HTTP_T_Status;
-			break;
-		case HTTP_HDR_RESPONSE:
-			htt = HTTP_T_Response;
-			break;
-		default:
-			htt = HTTP_T_Header;
-			break;
-		}
 		if (hp->hd[u].b >= hp->ws->s && hp->hd[u].e <= hp->ws->e) {
-			WSLH(w, htt, fd, hp, u);
+			WSLH(w, fd, hp, u);
 			continue;
 		}
-		l = hp->hd[u].e - hp->hd[u].b;
+		l = Tlen(hp->hd[u]);
 		p = WS_Alloc(hp->ws, l + 1);
 		if (p != NULL) {
-			WSLH(w, htt, fd, hp, u);
+			WSLH(w, fd, hp, u);
 			memcpy(p, hp->hd[u].b, l + 1);
 			hp->hd[u].b = p;
 			hp->hd[u].e = p + l;
 		} else {
-			WSLH(w, SLT_LostHeader, fd, hp, u);
+			WSLR(w, SLT_LostHeader, fd, hp->hd[u]);
 			hp->hd[u].b = NULL;
 			hp->hd[u].e = NULL;
 		}
@@ -901,10 +712,11 @@ http_SetHeader(struct worker *w, int fd, struct http *to, const char *hdr)
 /*--------------------------------------------------------------------*/
 
 static void
-http_PutField(struct worker *w, int fd, struct http *to, int field, const char *string)
+http_PutField(struct worker *w, int fd, struct http *to, int field,
+    const char *string)
 {
 	char *p;
-	int l;
+	unsigned l;
 
 	CHECK_OBJ_NOTNULL(to, HTTP_MAGIC);
 	l = strlen(string);
@@ -913,15 +725,18 @@ http_PutField(struct worker *w, int fd, struct http *to, int field, const char *
 		WSL(w, SLT_LostHeader, fd, "%s", string);
 		to->hd[field].b = NULL;
 		to->hd[field].e = NULL;
+		to->hdf[field] = 0;
 	} else {
 		memcpy(p, string, l + 1);
 		to->hd[field].b = p;
 		to->hd[field].e = p + l;
+		to->hdf[field] = 0;
 	}
 }
 
 void
-http_PutProtocol(struct worker *w, int fd, struct http *to, const char *protocol)
+http_PutProtocol(struct worker *w, int fd, struct http *to,
+    const char *protocol)
 {
 
 	http_PutField(w, fd, to, HTTP_HDR_PROTO, protocol);
@@ -935,17 +750,20 @@ http_PutStatus(struct worker *w, int fd, struct http *to, int status)
 	assert(status >= 0 && status <= 999);
 	sprintf(stat, "%d", status);
 	http_PutField(w, fd, to, HTTP_HDR_STATUS, stat);
+	to->status = status;
 }
 
 void
-http_PutResponse(struct worker *w, int fd, struct http *to, const char *response)
+http_PutResponse(struct worker *w, int fd, struct http *to,
+    const char *response)
 {
 
 	http_PutField(w, fd, to, HTTP_HDR_RESPONSE, response);
 }
 
 void
-http_PrintfHeader(struct worker *w, int fd, struct http *to, const char *fmt, ...)
+http_PrintfHeader(struct worker *w, int fd, struct http *to,
+    const char *fmt, ...)
 {
 	va_list ap;
 	unsigned l, n;
@@ -962,6 +780,7 @@ http_PrintfHeader(struct worker *w, int fd, struct http *to, const char *fmt, ..
 	} else {
 		to->hd[to->nhd].b = to->ws->f;
 		to->hd[to->nhd].e = to->ws->f + n;
+		to->hdf[to->nhd] = 0;
 		WS_Release(to->ws, n + 1);
 		to->nhd++;
 	}
@@ -974,10 +793,12 @@ http_Unset(struct http *hp, const char *hdr)
 	unsigned u, v;
 
 	for (v = u = HTTP_HDR_FIRST; u < hp->nhd; u++) {
-		if (http_IsHdr(&hp->hd[u], hdr)) 
+		if (http_IsHdr(&hp->hd[u], hdr))
 			continue;
-		if (v != u)
-			memcpy(&hp->hd[v], &hp->hd[u], sizeof hp->hd[v]);
+		if (v != u) {
+			memcpy(&hp->hd[v], &hp->hd[u], sizeof *hp->hd);
+			memcpy(&hp->hdf[v], &hp->hdf[u], sizeof *hp->hdf);
+		}
 		v++;
 	}
 	hp->nhd = v;
@@ -986,34 +807,34 @@ http_Unset(struct http *hp, const char *hdr)
 /*--------------------------------------------------------------------*/
 
 unsigned
-http_Write(struct worker *w, struct http *hp, int resp)
+http_Write(struct worker *w, const struct http *hp, int resp)
 {
 	unsigned u, l;
 
 	if (resp) {
 		AN(hp->hd[HTTP_HDR_STATUS].b);
-		l = WRK_WriteH(w, &hp->hd[HTTP_HDR_PROTO], " ");
-		WSLH(w, HTTP_T_Protocol, *w->wfd, hp, HTTP_HDR_PROTO);
-		l += WRK_WriteH(w, &hp->hd[HTTP_HDR_STATUS], " ");
-		WSLH(w, HTTP_T_Status, *w->wfd, hp, HTTP_HDR_STATUS);
-		l += WRK_WriteH(w, &hp->hd[HTTP_HDR_RESPONSE], "\r\n");
-		WSLH(w, HTTP_T_Response, *w->wfd, hp, HTTP_HDR_RESPONSE);
+		l = WRW_WriteH(w, &hp->hd[HTTP_HDR_PROTO], " ");
+		WSLH(w, *w->wfd, hp, HTTP_HDR_PROTO);
+		l += WRW_WriteH(w, &hp->hd[HTTP_HDR_STATUS], " ");
+		WSLH(w, *w->wfd, hp, HTTP_HDR_STATUS);
+		l += WRW_WriteH(w, &hp->hd[HTTP_HDR_RESPONSE], "\r\n");
+		WSLH(w, *w->wfd, hp, HTTP_HDR_RESPONSE);
 	} else {
 		AN(hp->hd[HTTP_HDR_URL].b);
-		l = WRK_WriteH(w, &hp->hd[HTTP_HDR_REQ], " ");
-		WSLH(w, HTTP_T_Request, *w->wfd, hp, HTTP_HDR_REQ);
-		l += WRK_WriteH(w, &hp->hd[HTTP_HDR_URL], " ");
-		WSLH(w, HTTP_T_URL, *w->wfd, hp, HTTP_HDR_URL);
-		l += WRK_WriteH(w, &hp->hd[HTTP_HDR_PROTO], "\r\n");
-		WSLH(w, HTTP_T_Protocol, *w->wfd, hp, HTTP_HDR_PROTO);
+		l = WRW_WriteH(w, &hp->hd[HTTP_HDR_REQ], " ");
+		WSLH(w, *w->wfd, hp, HTTP_HDR_REQ);
+		l += WRW_WriteH(w, &hp->hd[HTTP_HDR_URL], " ");
+		WSLH(w, *w->wfd, hp, HTTP_HDR_URL);
+		l += WRW_WriteH(w, &hp->hd[HTTP_HDR_PROTO], "\r\n");
+		WSLH(w, *w->wfd, hp, HTTP_HDR_PROTO);
 	}
 	for (u = HTTP_HDR_FIRST; u < hp->nhd; u++) {
 		AN(hp->hd[u].b);
 		AN(hp->hd[u].e);
-		l += WRK_WriteH(w, &hp->hd[u], "\r\n");
-		WSLH(w, HTTP_T_Header, *w->wfd, hp, u);
+		l += WRW_WriteH(w, &hp->hd[u], "\r\n");
+		WSLH(w, *w->wfd, hp, u);
 	}
-	l += WRK_Write(w, "\r\n", -1);
+	l += WRW_Write(w, "\r\n", -1);
 	return (l);
 }
 
@@ -1022,6 +843,7 @@ http_Write(struct worker *w, struct http *hp, int resp)
 void
 HTTP_Init(void)
 {
+
 #define HTTPH(a, b, c, d, e, f, g) b[0] = (char)strlen(b + 1);
 #include "http_headers.h"
 #undef HTTPH
