@@ -1,6 +1,6 @@
 /*-
  * Copyright (c) 2006 Verdens Gang AS
- * Copyright (c) 2006-2007 Linpro AS
+ * Copyright (c) 2006-2008 Linpro AS
  * All rights reserved.
  *
  * Author: Poul-Henning Kamp <phk@phk.freebsd.dk>
@@ -26,9 +26,11 @@
  * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
  * SUCH DAMAGE.
  *
- * $Id: cache_response.c 2034 2007-09-25 08:54:11Z des $
+ * $Id: cache_response.c 3701 2009-02-09 12:32:54Z tfheen $
  */
 
+#include "config.h"
+
 #include <sys/types.h>
 #include <sys/time.h>
 
@@ -36,40 +38,16 @@
 #include <stdlib.h>
 
 #include "shmlog.h"
-#include "heritage.h"
 #include "cache.h"
 
 /*--------------------------------------------------------------------*/
 
-void
-RES_Error(struct sess *sp, int code, const char *reason)
-{
-
-	/* get a pristine object */
-	HSH_Prealloc(sp);
-	sp->obj = sp->wrk->nobj;
-	sp->wrk->nobj = NULL;
-	sp->obj->busy = 1;
-
-	/* synthesize error page and send it */
-	SYN_ErrorPage(sp, code, reason, 0);
-	RES_BuildHttp(sp);
-	RES_WriteObj(sp);
-
-	/* GC the error page */
-	HSH_Unbusy(sp->obj);
-	HSH_Deref(sp->obj);
-	sp->obj = NULL;
-}
-
-/*--------------------------------------------------------------------*/
-
 static void
 res_do_304(struct sess *sp)
 {
 	char lm[64];
 
-	WSL(sp->wrk, SLT_Length, sp->fd, "%u", 0);
+	WSP(sp, SLT_Length, "%u", 0);
 
 	http_ClrHeader(sp->http);
 	sp->http->logtag = HTTP_Tx;
@@ -119,17 +97,21 @@ RES_BuildHttp(struct sess *sp)
 	if (sp->obj->response == 200 && sp->http->conds && res_do_conds(sp))
 		return;
 
-	WSL(sp->wrk, SLT_Length, sp->fd, "%u", sp->obj->len);
+	WSP(sp, SLT_Length, "%u", sp->obj->len);
 
 	http_ClrHeader(sp->http);
 	sp->http->logtag = HTTP_Tx;
-	http_CopyResp(sp->http, &sp->obj->http);
-	http_FilterFields(sp->wrk, sp->fd, sp->http, &sp->obj->http,
+	http_CopyResp(sp->http, sp->obj->http);
+	http_FilterFields(sp->wrk, sp->fd, sp->http, sp->obj->http,
 	    HTTPH_A_DELIVER);
-	
+
+	/* Only HTTP 1.1 can do Chunked encoding */
+	if (sp->http->protover < 1.1 && !VTAILQ_EMPTY(&sp->obj->esibits))
+		http_Unset(sp->http, H_Transfer_Encoding);
+
 	TIM_format(TIM_real(), time_str);
 	http_PrintfHeader(sp->wrk, sp->fd, sp->http, "Date: %s", time_str);
-	
+
 	if (sp->xid != sp->obj->xid)
 		http_PrintfHeader(sp->wrk, sp->fd, sp->http,
 		    "X-Varnish: %u %u", sp->xid, sp->obj->xid);
@@ -150,21 +132,37 @@ RES_WriteObj(struct sess *sp)
 {
 	struct storage *st;
 	unsigned u = 0;
+	char lenbuf[20];
 
 	CHECK_OBJ_NOTNULL(sp, SESS_MAGIC);
-
-	WRK_Reset(sp->wrk, &sp->fd);
-	sp->wrk->acct.hdrbytes += http_Write(sp->wrk, sp->http, 1);
 	CHECK_OBJ_NOTNULL(sp, SESS_MAGIC);
 
+	WRW_Reserve(sp->wrk, &sp->fd);
+
+	if (sp->esis == 0)
+		sp->acct_req.hdrbytes += http_Write(sp->wrk, sp->http, 1);
+
+	if (sp->wantbody && !VTAILQ_EMPTY(&sp->obj->esibits)) {
+		if (WRW_FlushRelease(sp->wrk)) {
+			vca_close_session(sp, "remote closed");
+			return;
+		}
+		ESI_Deliver(sp);
+		return;
+	}
+
 	if (sp->wantbody) {
+		if (sp->esis > 0 && sp->http->protover >= 1.1) {
+			sprintf(lenbuf, "%x\r\n", sp->obj->len);
+			(void)WRW_Write(sp->wrk, lenbuf, -1);
+		}
+
 		VTAILQ_FOREACH(st, &sp->obj->store, list) {
 			CHECK_OBJ_NOTNULL(sp, SESS_MAGIC);
 			CHECK_OBJ_NOTNULL(st, STORAGE_MAGIC);
-			AN(st->stevedore);
 			u += st->len;
-			sp->wrk->acct.bodybytes += st->len;
-#ifdef HAVE_SENDFILE
+			sp->acct_req.bodybytes += st->len;
+#ifdef SENDFILE_WORKS
 			/*
 			 * XXX: the overhead of setting up sendfile is not
 			 * XXX: epsilon and maybe not even delta, so avoid
@@ -174,16 +172,18 @@ RES_WriteObj(struct sess *sp)
 			if (st->fd >= 0 &&
 			    st->len >= params->sendfile_threshold) {
 				VSL_stats->n_objsendfile++;
-				WRK_Sendfile(sp->wrk, st->fd,
+				WRW_Sendfile(sp->wrk, st->fd,
 				    st->where, st->len);
 				continue;
 			}
-#endif /* HAVE_SENDFILE */
+#endif /* SENDFILE_WORKS */
 			VSL_stats->n_objwrite++;
-			WRK_Write(sp->wrk, st->ptr, st->len);
+			(void)WRW_Write(sp->wrk, st->ptr, st->len);
 		}
 		assert(u == sp->obj->len);
+		if (sp->esis > 0 && sp->http->protover >= 1.1)
+			(void)WRW_Write(sp->wrk, "\r\n", -1);
 	}
-	if (WRK_Flush(sp->wrk))
+	if (WRW_FlushRelease(sp->wrk))
 		vca_close_session(sp, "remote closed");
 }
