@@ -1,6 +1,6 @@
 /*-
  * Copyright (c) 2006 Verdens Gang AS
- * Copyright (c) 2006-2007 Linpro AS
+ * Copyright (c) 2006-2008 Linpro AS
  * All rights reserved.
  *
  * Author: Poul-Henning Kamp <phk@phk.freebsd.dk>
@@ -26,27 +26,34 @@
  * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
  * SUCH DAMAGE.
  *
- * $Id: cache_vrt.c 2139 2007-10-19 11:29:01Z des $
+ * $Id: cache_vrt.c 3726 2009-02-10 15:06:31Z tfheen $
  *
  * Runtime support for compiled VCL programs
  */
 
+#include "config.h"
+
 #include <sys/types.h>
 #include <sys/socket.h>
 
 #include <netinet/in.h>
+#include <arpa/inet.h>
 
 #include <stdio.h>
+#include <math.h>
 #include <string.h>
 #include <stdlib.h>
 #include <stdarg.h>
 
 #include "shmlog.h"
-#include "heritage.h"
 #include "vrt.h"
 #include "vrt_obj.h"
 #include "vcl.h"
 #include "cache.h"
+#include "hash_slinger.h"
+#include "cache_backend.h"
+
+void *vrt_magic_string_end = &vrt_magic_string_end;
 
 /*--------------------------------------------------------------------*/
 
@@ -56,26 +63,34 @@ VRT_error(struct sess *sp, unsigned code, const char *reason)
 
 	CHECK_OBJ_NOTNULL(sp, SESS_MAGIC);
 	WSL(sp->wrk, SLT_Debug, 0, "VCL_error(%u, %s)", code, reason);
-	sp->err_code = code;
-	sp->err_reason = reason;
+	sp->err_code = code ? code : 503;
+	sp->err_reason = reason ? reason : http_StatusMessage(sp->err_code);
 }
 
 /*--------------------------------------------------------------------*/
 
 void
-VRT_count(struct sess *sp, unsigned u)
+VRT_count(const struct sess *sp, unsigned u)
 {
 
 	CHECK_OBJ_NOTNULL(sp, SESS_MAGIC);
 	if (params->vcl_trace)
-		WSL(sp->wrk, SLT_VCL_trace, sp->fd, "%u %d.%d", u,
+		WSP(sp, SLT_VCL_trace, "%u %d.%d", u,
 		    sp->vcl->ref[u].line, sp->vcl->ref[u].pos);
 }
 
 /*--------------------------------------------------------------------*/
 
+void
+VRT_acl_log(const struct sess *sp, const char *msg)
+{
+	WSL(sp->wrk, SLT_VCL_acl, sp->fd, msg);
+}
+
+/*--------------------------------------------------------------------*/
+
 static struct http *
-vrt_selecthttp(struct sess *sp, enum gethdr_e where)
+vrt_selecthttp(const struct sess *sp, enum gethdr_e where)
 {
 	struct http *hp;
 
@@ -91,7 +106,7 @@ vrt_selecthttp(struct sess *sp, enum gethdr_e where)
 		hp = sp->http;
 		break;
 	case HDR_OBJ:
-		hp = &sp->obj->http;
+		hp = sp->obj->http;
 		break;
 	default:
 		INCOMPL();
@@ -101,7 +116,7 @@ vrt_selecthttp(struct sess *sp, enum gethdr_e where)
 }
 
 char *
-VRT_GetHdr(struct sess *sp, enum gethdr_e where, const char *n)
+VRT_GetHdr(const struct sess *sp, enum gethdr_e where, const char *n)
 {
 	char *p;
 	struct http *hp;
@@ -115,6 +130,7 @@ VRT_GetHdr(struct sess *sp, enum gethdr_e where, const char *n)
 
 /*--------------------------------------------------------------------*/
 
+/*lint -e{818} ap,hp could be const */
 static char *
 vrt_assemble_string(struct http *hp, const char *h, const char *p, va_list ap)
 {
@@ -129,17 +145,19 @@ vrt_assemble_string(struct http *hp, const char *h, const char *p, va_list ap)
 		if (b + x < e)
 			memcpy(b, h, x);
 		b += x;
-		if (b + 1 < e) 
+		if (b + 1 < e)
 			*b++ = ' ';
 	}
-	while (p != NULL) {
+	while (p != vrt_magic_string_end) {
+		if (p == NULL)
+			p = "(null)";
 		x = strlen(p);
 		if (b + x < e)
 			memcpy(b, p, x);
 		b += x;
 		p = va_arg(ap, const char *);
 	}
-	if (b + 1 < e) 
+	if (b + 1 < e)
 		*b++ = '\0';
 	if (b > e) {
 		WS_Release(hp->ws, 0);
@@ -155,7 +173,8 @@ vrt_assemble_string(struct http *hp, const char *h, const char *p, va_list ap)
 /*--------------------------------------------------------------------*/
 
 void
-VRT_SetHdr(struct sess *sp , enum gethdr_e where, const char *hdr, const char *p, ...)
+VRT_SetHdr(const struct sess *sp , enum gethdr_e where, const char *hdr,
+    const char *p, ...)
 {
 	struct http *hp;
 	va_list ap;
@@ -169,7 +188,7 @@ VRT_SetHdr(struct sess *sp , enum gethdr_e where, const char *hdr, const char *p
 	} else {
 		b = vrt_assemble_string(hp, hdr + 1, p, ap);
 		if (b == NULL) {
-			VSL(SLT_LostHeader, sp->fd, hdr + 1);
+			WSP(sp, SLT_LostHeader, "%s", hdr + 1);
 		} else {
 			http_Unset(hp, hdr);
 			http_SetHeader(sp->wrk, sp->fd, hp, b);
@@ -181,7 +200,8 @@ VRT_SetHdr(struct sess *sp , enum gethdr_e where, const char *hdr, const char *p
 /*--------------------------------------------------------------------*/
 
 static void
-vrt_do_string(struct worker *w, int fd, struct http *hp, int fld, const char *err, const char *p, va_list ap)
+vrt_do_string(struct worker *w, int fd, struct http *hp, int fld,
+    const char *err, const char *p, va_list ap)
 {
 	char *b;
 
@@ -198,7 +218,7 @@ vrt_do_string(struct worker *w, int fd, struct http *hp, int fld, const char *er
 
 #define VRT_DO_HDR(obj, hdr, http, fld)				\
 void								\
-VRT_l_##obj##_##hdr(struct sess *sp, const char *p, ...)	\
+VRT_l_##obj##_##hdr(const struct sess *sp, const char *p, ...)	\
 {								\
 	va_list ap;						\
 								\
@@ -207,6 +227,14 @@ VRT_l_##obj##_##hdr(struct sess *sp, const char *p, ...)	\
 	vrt_do_string(sp->wrk, sp->fd,				\
 	    http, fld, #obj "." #hdr, p, ap);			\
 	va_end(ap);						\
+}								\
+								\
+const char *							\
+VRT_r_##obj##_##hdr(const struct sess *sp)			\
+{								\
+	CHECK_OBJ_NOTNULL(sp, SESS_MAGIC);			\
+	CHECK_OBJ_NOTNULL(http, HTTP_MAGIC);			\
+	return (http->hd[fld].b);				\
 }
 
 VRT_DO_HDR(req,   request,	sp->http,		HTTP_HDR_REQ)
@@ -215,47 +243,104 @@ VRT_DO_HDR(req,   proto,	sp->http,		HTTP_HDR_PROTO)
 VRT_DO_HDR(bereq, request,	sp->bereq->http,	HTTP_HDR_REQ)
 VRT_DO_HDR(bereq, url,		sp->bereq->http,	HTTP_HDR_URL)
 VRT_DO_HDR(bereq, proto,	sp->bereq->http,	HTTP_HDR_PROTO)
-VRT_DO_HDR(obj,   proto,	&sp->obj->http,		HTTP_HDR_PROTO)
-VRT_DO_HDR(obj,   response,	&sp->obj->http,		HTTP_HDR_RESPONSE)
+VRT_DO_HDR(obj,   proto,	sp->obj->http,		HTTP_HDR_PROTO)
+VRT_DO_HDR(obj,   response,	sp->obj->http,		HTTP_HDR_RESPONSE)
 VRT_DO_HDR(resp,  proto,	sp->http,		HTTP_HDR_PROTO)
 VRT_DO_HDR(resp,  response,	sp->http,		HTTP_HDR_RESPONSE)
 
+/*--------------------------------------------------------------------*/
+
+/* XXX: review this */
+
 void
-VRT_l_obj_status(struct sess *sp, int num)
+VRT_l_obj_status(const struct sess *sp, int num)
 {
 	char *p;
 
 	assert(num >= 100 && num <= 999);
-	p = WS_Alloc(sp->obj->http.ws, 4);
+	p = WS_Alloc(sp->obj->http->ws, 4);
 	if (p == NULL)
-		WSL(sp->wrk, SLT_LostHeader, sp->fd, "obj.status");
+		WSP(sp, SLT_LostHeader, "%s", "obj.status");
 	else
 		sprintf(p, "%d", num);
-	http_SetH(&sp->obj->http, HTTP_HDR_STATUS, p);
+	http_SetH(sp->obj->http, HTTP_HDR_STATUS, p);
 }
 
 int
-VRT_r_obj_status(struct sess *sp)
+VRT_r_obj_status(const struct sess *sp)
 {
 	CHECK_OBJ_NOTNULL(sp, SESS_MAGIC);
 	CHECK_OBJ_NOTNULL(sp->obj, OBJECT_MAGIC);
-	return (atoi(sp->obj->http.hd[HTTP_HDR_STATUS].b));
+	/* XXX: use http_GetStatus() */
+	if (sp->obj->http->status)
+		return (sp->obj->http->status);
+	return (atoi(sp->obj->http->hd[HTTP_HDR_STATUS].b));
 }
 
 void
-VRT_l_resp_status(struct sess *sp, int num)
+VRT_l_resp_status(const struct sess *sp, int num)
 {
 	char *p;
 
 	assert(num >= 100 && num <= 999);
 	p = WS_Alloc(sp->http->ws, 4);
 	if (p == NULL)
-		WSL(sp->wrk, SLT_LostHeader, sp->fd, "resp.status");
+		WSP(sp, SLT_LostHeader, "%s", "resp.status");
 	else
 		sprintf(p, "%d", num);
 	http_SetH(sp->http, HTTP_HDR_STATUS, p);
 }
 
+int
+VRT_r_resp_status(const struct sess *sp)
+{
+	CHECK_OBJ_NOTNULL(sp, SESS_MAGIC);
+	CHECK_OBJ_NOTNULL(sp->http, HTTP_MAGIC);
+	return (atoi(sp->http->hd[HTTP_HDR_STATUS].b));
+}
+
+void
+VRT_l_bereq_connect_timeout(struct sess *sp, double num)
+{
+	CHECK_OBJ_NOTNULL(sp, SESS_MAGIC);
+	sp->connect_timeout = (num > 0 ? num : 0);
+}
+
+double
+VRT_r_bereq_connect_timeout(struct sess *sp)
+{
+	CHECK_OBJ_NOTNULL(sp, SESS_MAGIC);
+	return sp->connect_timeout;
+}
+
+void
+VRT_l_bereq_first_byte_timeout(struct sess *sp, double num)
+{
+	CHECK_OBJ_NOTNULL(sp, SESS_MAGIC);
+	sp->first_byte_timeout = (num > 0 ? num : 0);
+}
+
+double
+VRT_r_bereq_first_byte_timeout(struct sess *sp)
+{
+	CHECK_OBJ_NOTNULL(sp, SESS_MAGIC);
+	return sp->first_byte_timeout;
+}
+
+void
+VRT_l_bereq_between_bytes_timeout(struct sess *sp, double num)
+{
+	CHECK_OBJ_NOTNULL(sp, SESS_MAGIC);
+	sp->between_bytes_timeout = (num > 0 ? num : 0);
+}
+
+double
+VRT_r_bereq_between_bytes_timeout(struct sess *sp)
+{
+	CHECK_OBJ_NOTNULL(sp, SESS_MAGIC);
+	return sp->between_bytes_timeout;
+}
+
 /*--------------------------------------------------------------------*/
 
 void
@@ -263,63 +348,10 @@ VRT_handling(struct sess *sp, unsigned hand)
 {
 
 	CHECK_OBJ_NOTNULL(sp, SESS_MAGIC);
-	assert(!(hand & (hand -1)));	/* must be power of two */
+	assert(hand < VCL_RET_MAX);
 	sp->handling = hand;
 }
 
-/*--------------------------------------------------------------------*/
-
-void
-VRT_set_backend_name(struct backend *be, const char *p)
-{
-	CHECK_OBJ_NOTNULL(be, BACKEND_MAGIC);
-	be->vcl_name = p;
-}
-
-void
-VRT_alloc_backends(struct VCL_conf *cp)
-{
-	int i;
-
-	cp->backend = calloc(sizeof *cp->backend, cp->nbackend);
-	XXXAN(cp->backend);
-	for (i = 0; i < cp->nbackend; i++) {
-		cp->backend[i] = calloc(sizeof *cp->backend[i], 1);
-		XXXAN(cp->backend[i]);
-		cp->backend[i]->magic = BACKEND_MAGIC;
-		cp->backend[i]->dnsttl = 30;
-		VTAILQ_INIT(&cp->backend[i]->connlist);
-	}
-}
-
-void
-VRT_free_backends(struct VCL_conf *cp)
-{
-
-	(void)cp;	/* XXX */
-}
-
-void
-VRT_fini_backend(struct backend *be)
-{
-
-	(void)be;	/* XXX */
-}
-
-/*--------------------------------------------------------------------*/
-
-#define VBACKEND(type,onm,field)			\
-void							\
-VRT_l_backend_##onm(struct backend *be, type a)		\
-{							\
-	CHECK_OBJ_NOTNULL(be, BACKEND_MAGIC);		\
-	be->field = a;					\
-}							\
-
-VBACKEND(const char *,	host,	hostname)
-VBACKEND(const char *,	port,	portname)
-VBACKEND(double,	dnsttl,	dnsttl)
-
 /*--------------------------------------------------------------------
  * XXX: Working relative to t_req is maybe not the right thing, we could
  * XXX: have spent a long time talking to the backend since then.
@@ -332,33 +364,101 @@ VBACKEND(double,	dnsttl,	dnsttl)
  */
 
 void
-VRT_l_obj_ttl(struct sess *sp, double a)
+VRT_l_obj_ttl(const struct sess *sp, double a)
 {
 
 	CHECK_OBJ_NOTNULL(sp, SESS_MAGIC);
 	CHECK_OBJ_NOTNULL(sp->obj, OBJECT_MAGIC);	/* XXX */
-	WSL(sp->wrk, SLT_TTL, sp->fd, "%u VCL %.0f %.0f",
+	WSP(sp, SLT_TTL, "%u VCL %.0f %.0f",
 	    sp->obj->xid, a, sp->t_req);
-	if (a < 0)
-		a = 0;
-	sp->obj->ttl = sp->t_req + a;
-	if (sp->obj->heap_idx != 0)
-		EXP_TTLchange(sp->obj);
+	/*
+	 * If people set obj.ttl = 0s, they don't expect it to be cacheable
+	 * any longer, but it will still be for up to 1s - epsilon because
+	 * of the rounding to seconds.
+	 * We special case and make sure that rounding does not surprise.
+	 */
+	if (a <= 0)
+		sp->obj->ttl = sp->t_req - 1;
+	else
+		sp->obj->ttl = sp->t_req + a;
+	EXP_Rearm(sp->obj);
 }
 
 double
-VRT_r_obj_ttl(struct sess *sp)
+VRT_r_obj_ttl(const struct sess *sp)
 {
 	CHECK_OBJ_NOTNULL(sp, SESS_MAGIC);
 	CHECK_OBJ_NOTNULL(sp->obj, OBJECT_MAGIC);	/* XXX */
 	return (sp->obj->ttl - sp->t_req);
 }
 
+/*--------------------------------------------------------------------
+ * obj.grace is relative to obj.ttl, so no special magic is necessary.
+ */
+
+void
+VRT_l_obj_grace(const struct sess *sp, double a)
+{
+
+	CHECK_OBJ_NOTNULL(sp, SESS_MAGIC);
+	CHECK_OBJ_NOTNULL(sp->obj, OBJECT_MAGIC);	/* XXX */
+	if (a < 0)
+		a = 0;
+	sp->obj->grace = a;
+	EXP_Rearm(sp->obj);
+}
+
+double
+VRT_r_obj_grace(const struct sess *sp)
+{
+	CHECK_OBJ_NOTNULL(sp, SESS_MAGIC);
+	CHECK_OBJ_NOTNULL(sp->obj, OBJECT_MAGIC);	/* XXX */
+	if (isnan(sp->obj->grace))
+		return ((double)params->default_grace);
+	return (sp->obj->grace);
+}
+
+/*--------------------------------------------------------------------*/
+
+/* XXX: the VCL_info messages has unexpected fractions on the ttl */
+
+void
+VRT_l_obj_prefetch(const struct sess *sp, double a)
+{
+
+	CHECK_OBJ_NOTNULL(sp, SESS_MAGIC);
+	CHECK_OBJ_NOTNULL(sp->obj, OBJECT_MAGIC);	/* XXX */
+	sp->obj->prefetch = 0.0;
+	if (a == 0.0)
+		sp->obj->prefetch = a;
+	else if (a > 0.0 && a + sp->t_req <= sp->obj->ttl)
+		sp->obj->prefetch = a + sp->t_req;
+	else if (a < 0.0 && a + sp->obj->ttl > sp->t_req)
+		sp->obj->prefetch = a;
+	else if (a > 0.0)
+		WSL(sp->wrk, SLT_VCL_info, sp->id,
+		    "XID %u: obj.prefetch (%g) after TTL (%g), ignored.",
+		    sp->obj->xid, a, sp->obj->ttl - sp->t_req);
+	else /* if (a < 0.0) */
+		WSL(sp->wrk, SLT_VCL_info, sp->id,
+		    "XID %u: obj.prefetch (%g) less than ttl (%g), ignored.",
+		    sp->obj->xid, a, sp->obj->ttl - sp->t_req);
+	EXP_Rearm(sp->obj);
+}
+
+double
+VRT_r_obj_prefetch(const struct sess *sp)
+{
+	CHECK_OBJ_NOTNULL(sp, SESS_MAGIC);
+	CHECK_OBJ_NOTNULL(sp->obj, OBJECT_MAGIC);	/* XXX */
+	return (sp->obj->prefetch - sp->t_req);
+}
+
 /*--------------------------------------------------------------------*/
 
 #define VOBJ(type,onm,field)						\
 void									\
-VRT_l_obj_##onm(struct sess *sp, type a)				\
+VRT_l_obj_##onm(const struct sess *sp, type a)				\
 {									\
 	CHECK_OBJ_NOTNULL(sp, SESS_MAGIC);				\
 	CHECK_OBJ_NOTNULL(sp->obj, OBJECT_MAGIC);	/* XXX */	\
@@ -366,77 +466,88 @@ VRT_l_obj_##onm(struct sess *sp, type a)				\
 }									\
 									\
 type									\
-VRT_r_obj_##onm(struct sess *sp)					\
+VRT_r_obj_##onm(const struct sess *sp)					\
 {									\
 	CHECK_OBJ_NOTNULL(sp, SESS_MAGIC);				\
 	CHECK_OBJ_NOTNULL(sp->obj, OBJECT_MAGIC);	/* XXX */	\
 	return (sp->obj->field);					\
 }
 
-VOBJ(unsigned, valid, valid)
 VOBJ(unsigned, cacheable, cacheable)
 
 /*--------------------------------------------------------------------*/
 
 void
-VRT_l_req_backend(struct sess *sp, struct backend *be)
+VRT_l_req_backend(struct sess *sp, struct director *be)
 {
 	CHECK_OBJ_NOTNULL(sp, SESS_MAGIC);
-	sp->backend = be;
+	sp->director = be;
 }
 
-struct backend *
+/*lint -e{818} sp could be const */
+struct director *
 VRT_r_req_backend(struct sess *sp)
 {
 	CHECK_OBJ_NOTNULL(sp, SESS_MAGIC);
-	return (sp->backend);
+	return (sp->director);
 }
 
 /*--------------------------------------------------------------------*/
 
-#define  VREQ(n1, n2)					\
-const char *						\
-VRT_r_req_##n1(struct sess *sp)				\
-{							\
-	CHECK_OBJ_NOTNULL(sp, SESS_MAGIC);		\
-	CHECK_OBJ_NOTNULL(sp->http, HTTP_MAGIC);	\
-	return (sp->http->hd[n2].b);			\
-}
-
-VREQ(request, HTTP_HDR_REQ)
-VREQ(url, HTTP_HDR_URL)
-VREQ(proto, HTTP_HDR_PROTO)
-
-/*--------------------------------------------------------------------*/
-
-const char *
-VRT_r_resp_proto(struct sess *sp)
-{
-	CHECK_OBJ_NOTNULL(sp, SESS_MAGIC);
-	CHECK_OBJ_NOTNULL(sp->obj, OBJECT_MAGIC);
-	return (sp->obj->http.hd[HTTP_HDR_PROTO].b);
-}
-
-const char *
-VRT_r_resp_response(struct sess *sp)
-{
-	CHECK_OBJ_NOTNULL(sp, SESS_MAGIC);
-	CHECK_OBJ_NOTNULL(sp->obj, OBJECT_MAGIC);
-	return (sp->obj->http.hd[HTTP_HDR_RESPONSE].b);
-}
-
 int
-VRT_r_resp_status(struct sess *sp)
+VRT_r_req_restarts(const struct sess *sp)
+{
+
+	CHECK_OBJ_NOTNULL(sp, SESS_MAGIC);
+	return (sp->restarts);
+}
+
+/*--------------------------------------------------------------------
+ * req.grace
+ */
+
+void
+VRT_l_req_grace(struct sess *sp, double a)
+{
+
+	CHECK_OBJ_NOTNULL(sp, SESS_MAGIC);
+	if (a < 0)
+		a = 0;
+	sp->grace = a;
+}
+
+/*lint -e{818} sp could be const */
+double
+VRT_r_req_grace(struct sess *sp)
 {
 	CHECK_OBJ_NOTNULL(sp, SESS_MAGIC);
-	CHECK_OBJ_NOTNULL(sp->obj, OBJECT_MAGIC);
-	return (atoi(sp->obj->http.hd[HTTP_HDR_STATUS].b));
+	if (isnan(sp->grace))
+		return ((double)params->default_grace);
+	return (sp->grace);
+}
+
+/*--------------------------------------------------------------------
+ * req.xid
+ */
+
+/*lint -e{818} sp could be const */
+const char *
+VRT_r_req_xid(struct sess *sp)
+{
+	char *p;
+	int size;
+	CHECK_OBJ_NOTNULL(sp, SESS_MAGIC);
+
+	size = snprintf(NULL, 0, "%u", sp->xid) + 1;
+	AN(p = WS_Alloc(sp->http->ws, size));
+	assert(snprintf(p, size, "%u", sp->xid) < size);
+	return (p);
 }
 
 /*--------------------------------------------------------------------*/
 
 struct sockaddr *
-VRT_r_client_ip(struct sess *sp)
+VRT_r_client_ip(const struct sess *sp)
 {
 
 	return (sp->sockaddr);
@@ -447,11 +558,26 @@ VRT_r_server_ip(struct sess *sp)
 {
 
 	if (sp->mysockaddr->sa_family == AF_UNSPEC)
-		AZ(getsockname(sp->fd, sp->mysockaddr, &sp->mysockaddrlen));
+		assert(!getsockname(sp->fd, sp->mysockaddr, &sp->mysockaddrlen)
+		    || errno == ECONNRESET);
 
 	return (sp->mysockaddr);
 }
 
+int
+VRT_r_server_port(struct sess *sp)
+{
+	char abuf[TCP_ADDRBUFSIZE];
+	char pbuf[TCP_PORTBUFSIZE];
+
+	if (sp->mysockaddr->sa_family == AF_UNSPEC)
+		AZ(getsockname(sp->fd, sp->mysockaddr, &sp->mysockaddrlen));
+	TCP_name(sp->mysockaddr, sp->mysockaddrlen,
+	    abuf, sizeof abuf, pbuf, sizeof pbuf);
+
+	return (atoi(pbuf));
+}
+
 /*--------------------------------------------------------------------
  * Add an element to the array/list of hash bits.
  */
@@ -459,80 +585,263 @@ VRT_r_server_ip(struct sess *sp)
 void
 VRT_l_req_hash(struct sess *sp, const char *str)
 {
-	int l;
 
-	if (str == NULL)
-		str = "";
-	l = strlen(str);
-
-	/*
-	 * XXX: handle this by bouncing sp->vcl->nhashcount when it fails
-	 * XXX: and dispose of this request either by reallocating the
-	 * XXX: hashptr (if possible) or restarting/error the request
-	 */
-	xxxassert(sp->ihashptr < sp->nhashptr);
-
-	sp->hashptr[sp->ihashptr] = str;
-	sp->hashptr[sp->ihashptr + 1] = str + l;
-	sp->ihashptr += 2;
-	sp->lhashptr += l + 1;
+	HSH_AddString(sp, str);
 }
 
 /*--------------------------------------------------------------------*/
 
 double
-VRT_r_now(struct sess *sp)
+VRT_r_now(const struct sess *sp)
 {
 
 	(void)sp;
-	return (TIM_mono());
+	return (TIM_real());
 }
 
-double
-VRT_r_obj_lastuse(struct sess *sp)
+int
+VRT_r_obj_hits(const struct sess *sp)
 {
 
 	CHECK_OBJ_NOTNULL(sp, SESS_MAGIC);
 	CHECK_OBJ_NOTNULL(sp->obj, OBJECT_MAGIC);	/* XXX */
-	return (TIM_mono() - sp->obj->lru_stamp);
+	return (sp->obj->hits);
+}
+
+double
+VRT_r_obj_lastuse(const struct sess *sp)
+{
+
+	CHECK_OBJ_NOTNULL(sp, SESS_MAGIC);
+	CHECK_OBJ_NOTNULL(sp->obj, OBJECT_MAGIC);	/* XXX */
+	return (TIM_real() - sp->obj->last_use);
+}
+
+const char *
+VRT_r_obj_hash(const struct sess *sp)
+{
+
+	CHECK_OBJ_NOTNULL(sp, SESS_MAGIC);
+	if (sp->obj == NULL)
+		return (NULL);
+	CHECK_OBJ_NOTNULL(sp->obj, OBJECT_MAGIC);
+	if (sp->obj->objhead == NULL)
+		return (NULL);
+	CHECK_OBJ_NOTNULL(sp->obj->objhead, OBJHEAD_MAGIC);
+	return (sp->obj->objhead->hash);
+}
+
+unsigned
+VRT_r_req_backend_healthy(const struct sess *sp)
+{
+	CHECK_OBJ_NOTNULL(sp, SESS_MAGIC);
+	CHECK_OBJ_NOTNULL(sp->director, DIRECTOR_MAGIC);
+	return (sp->director->healthy(sp));
 }
 
 /*--------------------------------------------------------------------*/
 
 char *
-VRT_IP_string(struct sess *sp, struct sockaddr *sa)
+VRT_IP_string(const struct sess *sp, const struct sockaddr *sa)
 {
-	char h[64], p[8], *q;
-	socklen_t len = 0;
+	char *p;
+	const struct sockaddr_in *si4;
+	const struct sockaddr_in6 *si6;
+	const void *addr;
+	int len;
 
-	/* XXX can't rely on sockaddr.sa_len */
 	switch (sa->sa_family) {
 	case AF_INET:
-		len = sizeof(struct sockaddr_in);
+		len = INET_ADDRSTRLEN;
+		si4 = (const void *)sa;
+		addr = &(si4->sin_addr);
 		break;
 	case AF_INET6:
-		len = sizeof(struct sockaddr_in6);
+		len = INET6_ADDRSTRLEN;
+		si6 = (const void *)sa;
+		addr = &(si6->sin6_addr);
 		break;
 	default:
 		INCOMPL();
 	}
 	XXXAN(len);
-	TCP_name(sa, len, h, sizeof h, p, sizeof p);
-	q = WS_Alloc(sp->http->ws, strlen(h) + strlen(p) + 2);
-	AN(q);
-	strcpy(q, h);
-	strcat(q, ":");
-	strcat(q, p);
-	return (q);
+	AN(p = WS_Alloc(sp->http->ws, len));
+	AN(inet_ntop(sa->sa_family, addr, p, len));
+	return (p);
+}
+
+char *
+VRT_int_string(const struct sess *sp, int num)
+{
+	char *p;
+	int size;
+
+	size = snprintf(NULL, 0, "%d", num) + 1;
+	AN(p = WS_Alloc(sp->http->ws, size));
+	assert(snprintf(p, size, "%d", num) < size);
+	return (p);
+}
+
+char *
+VRT_double_string(const struct sess *sp, double num)
+{
+	char *p;
+	int size;
+
+	size = snprintf(NULL, 0, "%.3f", num) + 1;
+	AN(p = WS_Alloc(sp->http->ws, size));
+	assert(snprintf(p, size, "%.3f", num) < size);
+	return (p);
+}
+
+const char *
+VRT_backend_string(struct sess *sp)
+{
+	CHECK_OBJ_NOTNULL(sp, SESS_MAGIC);
+	if (sp->director == NULL)
+		return (NULL);
+	return (sp->director->vcl_name);
 }
 
 /*--------------------------------------------------------------------*/
 
 void
-VRT_purge(const char *regexp)
+VRT_Rollback(struct sess *sp)
 {
-	
-	AddBan(regexp);
+
+	*sp->http = *sp->http0;
+	WS_Reset(sp->ws, sp->ws_req);
+}
+
+/*--------------------------------------------------------------------*/
+
+/*lint -e{818} sp could be const */
+void
+VRT_panic(struct sess *sp, const char *str, ...)
+{
+	va_list ap;
+	char *b;
+
+	va_start(ap, str);
+	b = vrt_assemble_string(sp->http, "PANIC: ", str, ap);
+	va_end(ap);
+	lbv_assert("VCL", "", 0, b, 0, 2);
+}
+
+/*--------------------------------------------------------------------*/
+
+/*lint -e{818} sp could be const */
+void
+VRT_synth_page(struct sess *sp, unsigned flags, const char *str, ...)
+{
+	va_list ap;
+	const char *p;
+	struct vsb *vsb;
+
+	(void)flags;
+	CHECK_OBJ_NOTNULL(sp, SESS_MAGIC);
+	CHECK_OBJ_NOTNULL(sp->obj, OBJECT_MAGIC);
+	vsb = SMS_Makesynth(sp->obj);
+	AN(vsb);
+
+	vsb_cat(vsb, str);
+	va_start(ap, str);
+	p = va_arg(ap, const char *);
+	while (p != vrt_magic_string_end) {
+		if (p == NULL)
+			p = "(null)";
+		vsb_cat(vsb, p);
+		p = va_arg(ap, const char *);
+	}
+	va_end(ap);
+	SMS_Finish(sp->obj);
+	http_Unset(sp->obj->http, H_Content_Length);
+	http_PrintfHeader(sp->wrk, sp->fd, sp->obj->http,
+	    "Content-Length: %d", sp->obj->len);
+}
+
+/*--------------------------------------------------------------------*/
+
+void
+VRT_purge(struct sess *sp, char *cmds, ...)
+{
+	char *a1, *a2, *a3;
+	va_list ap;
+	struct ban *b;
+	int good;
+
+	(void)sp;
+	b = BAN_New();
+	va_start(ap, cmds);
+	a1 = cmds;
+	good = 0;
+	while (a1 != NULL) {
+		good = 0;
+		a2 = va_arg(ap, char *);
+		if (a2 == NULL)
+			break;
+		a3 = va_arg(ap, char *);
+		if (a3 == NULL)
+			break;
+		if (BAN_AddTest(NULL, b, a1, a2, a3))
+			break;
+		a1 = va_arg(ap, char *);
+		good = 1;
+	}
+	if (!good) 
+		/* XXX: report error how ? */
+		BAN_Free(b);
+	else
+		BAN_Insert(b);
+}
+
+/*--------------------------------------------------------------------*/
+
+void
+VRT_purge_string(struct sess *sp, char *str, ...)
+{
+	char *p, *a1, *a2, *a3;
+	char **av;
+	va_list ap;
+	struct ban *b;
+	int good;
+	int i;
+
+	va_start(ap, str);
+	p = vrt_assemble_string(sp->http, NULL, str, ap);
+	if (p == NULL)
+		/* XXX: report error how ? */
+		return;
+
+	av = ParseArgv(p, 0);
+	if (av[0] != NULL) {
+		/* XXX: report error how ? */
+		FreeArgv(av);
+		return;
+	}
+	b = BAN_New();
+	good = 0;
+	for (i = 1; ; i += 3) {
+		a1 = av[i];
+		if (a1 == NULL)
+			break;
+		good = 0;
+		a2 = av[i + 1];
+		if (a2 == NULL)
+			break;
+		a3 = av[i + 2];
+		if (a3 == NULL)
+			break;
+		if (BAN_AddTest(NULL, b, a1, a2, a3))
+			break;
+		good = 1;
+	}
+	if (!good) 
+		/* XXX: report error how ? */
+		BAN_Free(b);
+	else
+		BAN_Insert(b);
+	FreeArgv(av);
 }
 
 /*--------------------------------------------------------------------
@@ -546,3 +855,11 @@ VRT_strcmp(const char *s1, const char *s2)
 		return(1);
 	return (strcmp(s1, s2));
 }
+
+void
+VRT_memmove(void *dst, const void *src, unsigned len)
+{
+
+	(void)memmove(dst, src, len);
+}
+
