@@ -1,6 +1,6 @@
 /*-
  * Copyright (c) 2006 Verdens Gang AS
- * Copyright (c) 2006-2007 Linpro AS
+ * Copyright (c) 2006-2008 Linpro AS
  * All rights reserved.
  *
  * Author: Poul-Henning Kamp <phk@phk.freebsd.dk>
@@ -26,17 +26,21 @@
  * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
  * SUCH DAMAGE.
  *
- * $Id: mgt_cli.c 2307 2007-12-18 17:20:00Z des $
+ * $Id: mgt_cli.c 3732 2009-02-11 07:13:12Z tfheen $
  *
  * The management process' CLI handling
  */
 
+#include "config.h"
+
 #include <sys/types.h>
 
 #include <stdarg.h>
 #include <stdio.h>
+#include <ctype.h>
 #include <fcntl.h>
 #include <stdlib.h>
+#include <syslog.h>
 #include <string.h>
 #include <unistd.h>
 #include <sys/socket.h>
@@ -49,19 +53,33 @@
 #include "cli.h"
 #include "vsb.h"
 #include "cli_common.h"
+#include "heritage.h"
 #include "mgt.h"
 #include "mgt_cli.h"
-#include "mgt_event.h"
+#include "vev.h"
 #include "shmlog.h"
 
+#include "vlu.h"
 #include "vss.h"
 
 static int		cli_i = -1, cli_o = -1;
 
+struct telnet {
+	int			fd;
+	struct vev		*ev;
+	VTAILQ_ENTRY(telnet)	list;
+};
+
+static VTAILQ_HEAD(,telnet)	telnets = VTAILQ_HEAD_INITIALIZER(telnets);
+static void telnet_close_all(void);
+static void telnet_close_one(int fd);
+
+static int			dflag_copy;
+
 /*--------------------------------------------------------------------*/
 
 static void
-mcf_stats(struct cli *cli, char **av, void *priv)
+mcf_stats(struct cli *cli, const char * const *av, void *priv)
 {
 
 	(void)av;
@@ -74,75 +92,47 @@ mcf_stats(struct cli *cli, char **av, void *priv)
 #undef MAC_STAT
 }
 
-
-/*--------------------------------------------------------------------
- * Passthru of cli commands.  It is more or less just undoing what
- * the cli parser did, but such is life...
- */
+/*--------------------------------------------------------------------*/
 
 static void
-mcf_passthru(struct cli *cli, char **av, void *priv)
+mcf_help(struct cli *cli, const char * const *av, void *priv)
 {
-	struct vsb *sb;
-	char *p;
 	unsigned u;
-	int i;
+	char *p;
 
-	(void)priv;
-
-	/* Request */
-	if (cli_o <= 0) {
-		cli_result(cli, CLIS_CANT);
-		cli_out(cli, "Cache process not running");
-		return;
-	}
-	sb = vsb_new(NULL, NULL, 64, VSB_AUTOEXTEND);
-	XXXAN(sb);
-	for (u = 1; av[u] != NULL; u++) {
-		if (u > 1)
-			vsb_putc(sb, ' ');
-		vsb_putc(sb, '"');
-		for (p = av[u]; *p; p++) {
-			switch (*p) {
-			case '\\':
-				vsb_cat(sb, "\\\\");
-				break;
-			case '\n':
-				vsb_cat(sb, "\\n");
-				break;
-			case '"':
-				vsb_cat(sb, "\\\"");
-				break;
-			default:
-				vsb_putc(sb, *p);
-			}
+	cli_func_help(cli, av, priv);
+	if (cli_o >= 0 && (av[2] == NULL || *av[2] == '-')) {
+		p = NULL;
+		if (!mgt_cli_askchild(&u, &p,
+		    "help %s\n", av[2] != NULL ? av[2] : "")) {
+			cli_out(cli, "%s", p);
+			cli_result(cli, u);
 		}
-		vsb_putc(sb, '"');
+		free(p);
 	}
-	vsb_putc(sb, '\n');
-	xxxassert(!vsb_overflowed(sb));
-	vsb_finish(sb);
-	i = write(cli_o, vsb_data(sb), vsb_len(sb));
-	xxxassert(i == vsb_len(sb));
-	vsb_delete(sb);
-
-	i = cli_readres(cli_i, &u, &p, 3.0);
-	cli_result(cli, u);
-	cli_out(cli, "%s", p);
-	free(p);
-
 }
 
 /*--------------------------------------------------------------------*/
 
-static struct cli_proto *cli_proto;
+static void
+mcf_close(struct cli *cli, const char *const *av, void *priv)
+{
+
+	(void)av;
+	(void)priv;
+	cli_out(cli, "Closing CLI connection");
+	cli_result(cli, CLIS_CLOSE);
+}
+
+/*--------------------------------------------------------------------*/
 
 /* XXX: what order should this list be in ? */
-static struct cli_proto mgt_cli_proto[] = {
+static struct cli_proto cli_proto[] = {
+	{ CLI_HELP,		mcf_help, cli_proto },
 	{ CLI_PING,		cli_func_ping },
 	{ CLI_SERVER_STATUS,	mcf_server_status, NULL },
 	{ CLI_SERVER_START,	mcf_server_startstop, NULL },
-	{ CLI_SERVER_STOP,	mcf_server_startstop, &cli_proto },
+	{ CLI_SERVER_STOP,	mcf_server_startstop, cli_proto },
 	{ CLI_STATS,		mcf_stats, NULL },
 	{ CLI_VCL_LOAD,		mcf_config_load, NULL },
 	{ CLI_VCL_INLINE,	mcf_config_inline, NULL },
@@ -152,71 +142,23 @@ static struct cli_proto mgt_cli_proto[] = {
 	{ CLI_VCL_SHOW,		mcf_config_show, NULL },
 	{ CLI_PARAM_SHOW,	mcf_param_show, NULL },
 	{ CLI_PARAM_SET,	mcf_param_set, NULL },
-	{ CLI_HELP,		cli_func_help, NULL },
+
+	{ CLI_QUIT,		mcf_close, NULL},
 #if 0
 	{ CLI_SERVER_RESTART },
 	{ CLI_ZERO },
 	{ CLI_VERBOSE,		m_cli_func_verbose, NULL },
-	{ CLI_EXIT, 		m_cli_func_exit, NULL},
-	{ CLI_QUIT },
-	{ CLI_BYE },
 #endif
 	{ NULL }
 };
 
-
-/*--------------------------------------------------------------------*/
-
-void
-mgt_cli_init(void)
-{
-	struct cli_proto *cp;
-	unsigned u, v;
-
-	/*
-	 * Build the joint cli_proto by combining the manager process
-	 * entries with with the cache process entries.  The latter
-	 * get a "passthough" function in the joint list
-	 */
-	u = 0;
-	for (cp = mgt_cli_proto; cp->request != NULL; cp++)
-		u++;
-	for (cp = CLI_cmds; cp->request != NULL; cp++)
-		u++;
-	cli_proto = calloc(sizeof *cli_proto, u + 1);
-	XXXAN(cli_proto);
-	u = 0;
-	for (cp = mgt_cli_proto; cp->request != NULL; cp++)
-		cli_proto[u++] = *cp;
-	for (cp = CLI_cmds; cp->request != NULL; cp++) {
-		/* Skip any cache commands we already have in the manager */
-		for (v = 0; v < u; v++)
-			if (!strcmp(cli_proto[v].request, cp->request))
-				break;
-		if (v < u)
-			continue;
-		cli_proto[u] = *cp;
-		cli_proto[u].func = mcf_passthru;
-		u++;
-	}
-
-	/* Fixup the entry for 'help' entry */
-	for (u = 0; cli_proto[u].request != NULL; u++) {
-		if (!strcmp(cli_proto[u].request, "help")) {
-			cli_proto[u].priv = cli_proto;
-			break;
-		}
-	}
-}
-
 /*--------------------------------------------------------------------
  * Ask the child something over CLI, return zero only if everything is
  * happy happy.
  */
 
 int
-mgt_cli_askchild(unsigned *status, char **resp, const char *fmt, ...)
-{
+mgt_cli_askchild(unsigned *status, char **resp, const char *fmt, ...) {
 	char *p;
 	int i, j;
 	va_list ap;
@@ -247,7 +189,7 @@ mgt_cli_askchild(unsigned *status, char **resp, const char *fmt, ...)
 		return (CLIS_COMMS);
 	}
 
-	i = cli_readres(cli_i, &u, resp, 5.0);
+	(void)cli_readres(cli_i, &u, resp, params->cli_timeout);
 	if (status != NULL)
 		*status = u;
 	return (u == CLIS_OK ? 0 : u);
@@ -279,104 +221,155 @@ mgt_cli_stop_child(void)
 struct cli_port {
 	unsigned		magic;
 #define CLI_PORT_MAGIC		0x5791079f
-	struct ev		*ev;
+	struct vev		*ev;
 	int			fdi;
 	int			fdo;
 	int			verbose;
-	char			*buf;
-	int			nbuf; /* next free position in buf */
-	int			lbuf; /* length of buf */
+	struct vlu		*vlu;
 	struct cli		cli[1];
-	char			name[30];
+	char			*name;
 };
 
 static int
-mgt_cli_callback(struct ev *e, int what)
+mgt_cli_vlu(void *priv, const char *p)
 {
 	struct cli_port *cp;
-	char *p, *q;
+	char *q;
+	unsigned u;
 	int i;
 
-	CAST_OBJ_NOTNULL(cp, e->priv, CLI_PORT_MAGIC);
+	CAST_OBJ_NOTNULL(cp, priv, CLI_PORT_MAGIC);
+	vsb_clear(cp->cli->sb);
 
-	if (what & (EV_ERR | EV_HUP | EV_GONE))
-		goto cli_close;
+	/* Skip whitespace */
+	for (; isspace(*p); p++)
+		continue;
 
-	/* grow the buffer if it is full */
-	if (cp->nbuf == cp->lbuf) {
-		cp->lbuf += cp->lbuf;
-		cp->buf = realloc(cp->buf, cp->lbuf);
-		XXXAN(cp->buf);
-	}
+	/* Ignore empty lines */
+	if (*p == '\0')
+		return (0);
 
-	/* read more data into the buffer */
-	i = read(cp->fdi, cp->buf + cp->nbuf, cp->lbuf - cp->nbuf);
-	if (i <= 0)
-		goto cli_close;
-	cp->nbuf += i;
-
-	for (p = q = cp->buf; q < cp->buf + cp->nbuf; ++q) {
-		if (*q != '\n')
-			continue;
-		/* got a newline == got a command */
-		*q = '\0';
+	cli_dispatch(cp->cli, cli_proto, p);
+	vsb_finish(cp->cli->sb);
+	AZ(vsb_overflowed(cp->cli->sb));
+	if (cp->cli->result == CLIS_UNKNOWN) {
+		/*
+		 * Command not recognized in master, try cacher if it is
+		 * running.
+		 */
 		vsb_clear(cp->cli->sb);
-		cli_dispatch(cp->cli, cli_proto, p);
+		cp->cli->result = CLIS_OK;
+		if (cli_o <= 0) {
+			cli_result(cp->cli, CLIS_UNKNOWN);
+			cli_out(cp->cli,
+			    "Unknown request in manager process "
+			    "(child not running).\n"
+			    "Type 'help' for more info.");
+		} else {
+			i = write(cli_o, p, strlen(p));
+			xxxassert(i == strlen(p));
+			i = write(cli_o, "\n", 1);
+			xxxassert(i == 1);
+			(void)cli_readres(cli_i, &u, &q, params->cli_timeout);
+			cli_result(cp->cli, u);
+			cli_out(cp->cli, "%s", q);
+			free(q);
+		}
 		vsb_finish(cp->cli->sb);
-
-		/* send the result back */
-		if (cli_writeres(cp->fdo, cp->cli))
-			goto cli_close;
-
-		/* ready for next command */
-		p = q + 1;
+		AZ(vsb_overflowed(cp->cli->sb));
 	}
 
-	/* see if there's any data left in the buffer */
-	if (p != q) {
-		assert(q == cp->buf + cp->nbuf);
-		cp->nbuf -= (p - cp->buf);
-		memmove(cp->buf, p, cp->nbuf);
-	} else
-		cp->nbuf = 0;
+	/* send the result back */
+	syslog(LOG_INFO, "CLI %d result %d \"%s\"",
+	    cp->fdi, cp->cli->result, p);
+	if (cli_writeres(cp->fdo, cp->cli) || cp->cli->result == CLIS_CLOSE)
+		return (1);
 	return (0);
+}
 
-cli_close:
+/*--------------------------------------------------------------------
+ * Get rid of a CLI session.
+ *
+ * Always and only called from mgt_cli_callback().
+ *
+ * We must get rid of everything but the event, which gets GC'ed by
+ * ev_schdule_one() when mgt_cli_callback, through our return value
+ * returns non-zero.
+ */
+
+static int
+mgt_cli_close(struct cli_port *cp)
+{
+
+	CHECK_OBJ_NOTNULL(cp, CLI_PORT_MAGIC);
+	syslog(LOG_NOTICE, "CLI %d closed", cp->fdi);
+	free(cp->name);
 	vsb_delete(cp->cli->sb);
-	free(cp->buf);
-	close(cp->fdi);
-	if (cp->fdi == 0)
-		open("/dev/null", O_RDONLY);
-	close(cp->fdo);
-	if (cp->fdo == 1) {
-		close(2);
-		open("/dev/null", O_WRONLY);
-		open("/dev/null", O_WRONLY);
+	VLU_Destroy(cp->vlu);
+
+	(void)close(cp->fdi);
+	if (cp->fdo != cp->fdi)
+		(void)close(cp->fdo);
+
+	/* Special case for stdin/out/err */
+	if (cp->fdi == 0) {
+		assert(open("/dev/null", O_RDONLY) == 0);
+		assert(open("/dev/null", O_WRONLY) == 1);
+		(void)close(2);
+		assert(open("/dev/null", O_WRONLY) == 2);
+
+		if (dflag_copy == 2) {
+			mgt_stop_child();
+			telnet_close_all();
+		}
+	} else {
+		telnet_close_one(cp->fdi);
 	}
+
 	free(cp);
 	return (1);
 }
 
+/*--------------------------------------------------------------------
+ * Callback whenever something happens to the input fd of the session.
+ */
+
+static int
+mgt_cli_callback(const struct vev *e, int what)
+{
+	struct cli_port *cp;
+
+	CAST_OBJ_NOTNULL(cp, e->priv, CLI_PORT_MAGIC);
+
+	if (what & (EV_ERR | EV_HUP | EV_GONE))
+		return (mgt_cli_close(cp));
+
+	if (VLU_Fd(cp->fdi, cp->vlu))
+		return (mgt_cli_close(cp));
+	return (0);
+}
+
+/*--------------------------------------------------------------------*/
+
 void
-mgt_cli_setup(int fdi, int fdo, int verbose)
+mgt_cli_setup(int fdi, int fdo, int verbose, const char *ident)
 {
 	struct cli_port *cp;
 
 	cp = calloc(sizeof *cp, 1);
 	XXXAN(cp);
+	cp->vlu = VLU_New(cp, mgt_cli_vlu, params->cli_buffer);
 
-	sprintf(cp->name, "cli %d->%d", fdi, fdo);
+	cp->name = strdup(ident);
+	XXXAN(cp->name);
+	syslog(LOG_NOTICE, "CLI %d open from %s", fdi, cp->name);
 	cp->magic = CLI_PORT_MAGIC;
 
 	cp->fdi = fdi;
 	cp->fdo = fdo;
 	cp->verbose = verbose;
 
-	cp->lbuf = 4096;
-	cp->buf = malloc(cp->lbuf);
-	XXXAN(cp->buf);
-
-	cp->cli->sb = vsb_new(NULL, NULL, 0, VSB_AUTOEXTEND);
+	cp->cli->sb = vsb_newauto();
 	XXXAN(cp->cli->sb);
 
 	cp->ev = calloc(sizeof *cp->ev, 1);
@@ -385,52 +378,120 @@ mgt_cli_setup(int fdi, int fdo, int verbose)
 	cp->ev->fd_flags = EV_RD;
 	cp->ev->callback = mgt_cli_callback;
 	cp->ev->priv = cp;
-	ev_add(mgt_evb, cp->ev);
+	AZ(vev_add(mgt_evb, cp->ev));
+}
+
+/*--------------------------------------------------------------------*/
+
+static void
+telnet_close_one(int fd)
+{
+	struct telnet *tn, *tn2;
+
+	VTAILQ_FOREACH_SAFE(tn, &telnets, list, tn2) {
+		if (tn->fd != fd)
+			continue;
+		VTAILQ_REMOVE(&telnets, tn, list);
+		(void)close(tn->fd);
+		free(tn);
+		break;
+	}
+}
+
+
+static void
+telnet_close_all()
+{
+	struct telnet *tn, *tn2;
+
+	VTAILQ_FOREACH_SAFE(tn, &telnets, list, tn2) {
+		VTAILQ_REMOVE(&telnets, tn, list);
+		AZ(close(tn->fd));
+		free(tn);
+	}
+}
+
+static struct telnet *
+telnet_new(int fd)
+{
+	struct telnet *tn;
+
+	tn = calloc(sizeof *tn, 1);
+	AN(tn);
+	tn->fd = fd;
+	VTAILQ_INSERT_TAIL(&telnets, tn, list);
+	return (tn);
 }
 
 static int
-telnet_accept(struct ev *ev, int what)
+telnet_accept(const struct vev *ev, int what)
 {
 	struct sockaddr_storage addr;
 	socklen_t addrlen;
 	int i;
+	char abuf1[TCP_ADDRBUFSIZE], abuf2[TCP_ADDRBUFSIZE];
+	char pbuf1[TCP_PORTBUFSIZE], pbuf2[TCP_PORTBUFSIZE];
+	char *p;
 
 	(void)what;
 	addrlen = sizeof addr;
 	i = accept(ev->fd, (void *)&addr, &addrlen);
+	if (i < 0 && errno == EBADF)
+		return (1);
 	if (i < 0)
 		return (0);
 
-	mgt_cli_setup(i, i, 0);
+	TCP_myname(ev->fd, abuf1, sizeof abuf1, pbuf1, sizeof pbuf1);
+	TCP_name((void*)&addr, addrlen, abuf2, sizeof abuf2,
+	    pbuf2, sizeof pbuf2);
+	asprintf(&p, "telnet %s:%s %s:%s", abuf2, pbuf2, abuf1, pbuf1);
+	XXXAN(p);
+
+	(void)telnet_new(i);
+
+	mgt_cli_setup(i, i, 0, p);
+	free(p);
 	return (0);
 }
 
-int
-mgt_cli_telnet(const char *T_arg)
+void
+mgt_cli_telnet(int dflag, const char *T_arg)
 {
 	struct vss_addr **ta;
 	char *addr, *port;
-	int i, n;
+	int i, n, sock, good;
+	struct telnet *tn;
+
+	dflag_copy = dflag;
 
 	XXXAZ(VSS_parse(T_arg, &addr, &port));
-	XXXAN(n = VSS_resolve(addr, port, &ta));
-	free(addr);
-	free(port);
+	n = VSS_resolve(addr, port, &ta);
 	if (n == 0) {
 		fprintf(stderr, "Could not open management port\n");
 		exit(2);
 	}
+	good = 0;
 	for (i = 0; i < n; ++i) {
-		int sock = VSS_listen(ta[i], 1);
-		struct ev *ev = ev_new();
-		XXXAN(ev);
-		ev->fd = sock;
-		ev->fd_flags = POLLIN;
-		ev->callback = telnet_accept;
-		ev_add(mgt_evb, ev);
+		sock = VSS_listen(ta[i], 10);
+		if (sock < 0)
+			continue;
+		good++;
+		tn = telnet_new(sock);
+		tn->ev = vev_new();
+		XXXAN(tn->ev);
+		tn->ev->fd = sock;
+		tn->ev->fd_flags = POLLIN;
+		tn->ev->callback = telnet_accept;
+		AZ(vev_add(mgt_evb, tn->ev));
 		free(ta[i]);
 		ta[i] = NULL;
 	}
 	free(ta);
-	return (0);
+	if (good == 0) {
+		REPORT(LOG_ERR, "-T %s:%s could not be listened on.",
+		    addr, port);
+		exit(2);
+	}
+	free(addr);
+	free(port);
 }
