@@ -1,6 +1,6 @@
 /*-
  * Copyright (c) 2006 Verdens Gang AS
- * Copyright (c) 2006-2007 Linpro AS
+ * Copyright (c) 2006-2008 Linpro AS
  * All rights reserved.
  *
  * Author: Poul-Henning Kamp <phk@phk.freebsd.dk>
@@ -26,11 +26,13 @@
  * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
  * SUCH DAMAGE.
  *
- * $Id: storage_file.c 2239 2007-11-08 21:30:05Z des $
+ * $Id: storage_file.c 3667 2009-02-06 13:25:02Z tfheen $
  *
  * Storage method based on mmap'ed file
  */
 
+#include "config.h"
+
 #include <sys/param.h>
 #include <sys/mman.h>
 #include <sys/socket.h>
@@ -60,8 +62,10 @@
 #ifndef HAVE_ASPRINTF
 #include "compat/asprintf.h"
 #endif
+
 #include "shmlog.h"
 #include "cache.h"
+#include "stevedore.h"
 
 #ifndef MAP_NOCORE
 #define MAP_NOCORE 0 /* XXX Linux */
@@ -104,32 +108,50 @@ struct smf {
 };
 
 struct smf_sc {
-	char			*filename;
+	const char		*filename;
 	int			fd;
 	unsigned		pagesize;
 	uintmax_t		filesize;
 	struct smfhead		order;
 	struct smfhead		free[NBUCKET];
 	struct smfhead		used;
-	MTX			mtx;
+	struct lock		mtx;
 };
 
+/*--------------------------------------------------------------------*/
+
+static uintmax_t
+smf_fsspace(int fd, unsigned *bs)
+{
+#if defined(HAVE_SYS_STATVFS_H)
+	struct statvfs fsst;
+
+	AZ(fstatvfs(fd, &fsst));
+#elif defined(HAVE_SYS_MOUNT_H) || defined(HAVE_SYS_VFS_H)
+	struct statfs fsst;
+
+	AZ(fstatfs(sc->fd, &fsst));
+#else
+#error no struct statfs / struct statvfs
+#endif
+
+	/* We use units of the larger of filesystem blocksize and pagesize */
+	if (*bs < fsst.f_bsize)
+		*bs = fsst.f_bsize;
+	xxxassert(*bs % fsst.f_bsize == 0);
+	return (fsst.f_bsize * fsst.f_bavail);
+}
+
+
 /*--------------------------------------------------------------------*/
 
 static void
 smf_calcsize(struct smf_sc *sc, const char *size, int newfile)
 {
-#if defined(HAVE_SYS_STATVFS_H)
-	struct statvfs fsst;
-#elif defined(HAVE_SYS_MOUNT_H) || defined(HAVE_SYS_VFS_H)
-	struct statfs fsst;
-#else
-#error no struct statfs / struct statvfs
-#endif
 	uintmax_t l, fssize;
 	unsigned bs;
-	char suff[2];
-	int i, explicit;
+	const char *q;
+	int i;
 	off_t o;
 	struct stat st;
 
@@ -137,111 +159,66 @@ smf_calcsize(struct smf_sc *sc, const char *size, int newfile)
 	AZ(fstat(sc->fd, &st));
 	xxxassert(S_ISREG(st.st_mode));
 
-#if defined(HAVE_SYS_STATVFS_H)
-	AZ(fstatvfs(sc->fd, &fsst));
-#elif defined(HAVE_SYS_MOUNT_H) || defined(HAVE_SYS_VFS_H)
-	AZ(fstatfs(sc->fd, &fsst));
-#else
-#error no struct statfs / struct statvfs
-#endif
-
-	/* We use units of the larger of filesystem blocksize and pagesize */
 	bs = sc->pagesize;
-	if (bs < fsst.f_bsize)
-		bs = fsst.f_bsize;
+	fssize = smf_fsspace(sc->fd, &bs);
 	xxxassert(bs % sc->pagesize == 0);
-	xxxassert(bs % fsst.f_bsize == 0);
-	fssize = fsst.f_bsize * fsst.f_bavail;
-
-	i = sscanf(size, "%ju%1s", &l, suff); /* can return -1, 0, 1 or 2 */
-
-	explicit = i;
-	if (i == 0) {
-		fprintf(stderr,
-		    "Error: (-sfile) size \"%s\" not understood\n", size);
-		exit (2);
-	}
-
-	if (i >= 1 && l == 0) {
-		fprintf(stderr,
-		    "Error: (-sfile) zero size not permitted\n");
-		exit (2);
-	}
-
-	if (i == -1 && !newfile) /* Use the existing size of the file */
-		l = st.st_size;
-
-	/* We must have at least one block */
-	if (l < bs) {
-		if (i == -1) {
-			fprintf(stderr,
-			    "Info: (-sfile) default to 80%% size\n");
-			l = 80;
-			suff[0] = '%';
-			i = 2;
-		}
-
-		if (i == 2) {
-			if (suff[0] == 'k' || suff[0] == 'K')
-				l *= 1024UL;
-			else if (suff[0] == 'm' || suff[0] == 'M')
-				l *= 1024UL * 1024UL;
-			else if (suff[0] == 'g' || suff[0] == 'G')
-				l *= 1024UL * 1024UL * 1024UL;
-			else if (suff[0] == 't' || suff[0] == 'T')
-				l *= (uintmax_t)(1024UL * 1024UL) *
-				    (uintmax_t)(1024UL * 1024UL);
-			else if (suff[0] == '%') {
-				l *= fssize;
-				l /= 100;
-			}
-		}
 
+	if ((size == NULL || *size == '\0') && !newfile) {
 		/*
-		 * This trickery wouldn't be necessary if X/Open would
-		 * just add OFF_MAX to <limits.h>...
+		 * We have no size specification, but an existing file,
+		 * use it's existing size.
 		 */
-		o = l;
-		if (o != l || o < 0) {
-			do {
-				l >>= 1;
-				o = l;
-			} while (o != l || o < 0);
-			fprintf(stderr, "WARNING: storage file size reduced"
-			    " to %ju due to system limitations\n", l);
-		}
+		l = st.st_size;
+	} else {
+		AN(size);
+		q = str2bytes(size, &l, fssize);
 
-		if (l < st.st_size) {
-			AZ(ftruncate(sc->fd, l));
-		} else if (l - st.st_size > fssize) {
-			l = fssize * 80 / 100;
-			fprintf(stderr, "WARNING: storage file size reduced"
-			    " to %ju (80%% of available disk space)\n", l);
-		}
+		if (q != NULL)
+			ARGV_ERR("(-sfile) size \"%s\": %s\n", size, q);
+	}
+
+	/*
+	 * This trickery wouldn't be necessary if X/Open would
+	 * just add OFF_MAX to <limits.h>...
+	 */
+	i = 0;
+	while(1) {
+		o = l;
+		if (o == l && o > 0)
+			break;
+		l >>= 1;
+		i++;
+	}
+	if (i)
+		fprintf(stderr, "WARNING: storage file size reduced"
+		    " to %ju due to system limitations\n", l);
+
+	if (l < st.st_size) {
+		AZ(ftruncate(sc->fd, (off_t)l));
+	} else if (l - st.st_size > fssize) {
+		l = fssize * 80 / 100;
+		fprintf(stderr, "WARNING: storage file size reduced"
+		    " to %ju (80%% of available disk space)\n", l);
 	}
 
 	/* round down to multiple of filesystem blocksize or pagesize */
 	l -= (l % bs);
 
-	if (l < MINPAGES * (uintmax_t)sc->pagesize) {
-		fprintf(stderr,
-		    "Error: size too small, at least %ju needed\n",
+	if (l < MINPAGES * (uintmax_t)sc->pagesize)
+		ARGV_ERR("size too small, at least %ju needed\n",
 		    (uintmax_t)MINPAGES * sc->pagesize);
-		exit (2);
-	}
 
-	if (explicit < 3 && sizeof(void *) == 4 && l > INT32_MAX) {
+	if (sizeof(void *) == 4 && l > INT32_MAX) { /*lint !e506 !e774 */
 		fprintf(stderr,
-		    "NB: Limiting size to 2GB on 32 bit architecture to"
-		    " prevent running out of\naddress space."
-		    "  Specifiy explicit size to override.\n"
+		    "NB: Storage size limited to 2GB on 32 bit architecture,\n"
+		    "NB: otherwise we could run out of address space.\n"
 		);
 		l = INT32_MAX;
 		l -= (l % bs);
 	}
 
-	printf("file %s size %ju bytes (%ju fs-blocks, %ju pages)\n",
-	    sc->filename, l, l / fsst.f_bsize, l / sc->pagesize);
+	printf("storage_file: filename: %s size %ju MB.\n",
+	    sc->filename, l / (1024 * 1024));
 
 	sc->filesize = l;
 }
@@ -251,19 +228,45 @@ smf_initfile(struct smf_sc *sc, const char *size, int newfile)
 {
 	smf_calcsize(sc, size, newfile);
 
-	AZ(ftruncate(sc->fd, sc->filesize));
+	AZ(ftruncate(sc->fd, (off_t)sc->filesize));
 
 	/* XXX: force block allocation here or in open ? */
 }
 
+static const char default_size[] = "50%";
+static const char default_filename[] = ".";
+
 static void
-smf_init(struct stevedore *parent, const char *spec)
+smf_init(struct stevedore *parent, int ac, char * const *av)
 {
-	char *size;
-	char *p, *q;
+	const char *size, *fn, *r;
+	char *q, *p;
 	struct stat st;
 	struct smf_sc *sc;
 	unsigned u;
+	uintmax_t page_size;
+
+	AZ(av[ac]);
+
+	fn = default_filename;
+	size = default_size;
+	page_size = getpagesize();
+
+	if (ac > 3)
+		ARGV_ERR("(-sfile) too many arguments\n");
+	if (ac > 0 && *av[0] != '\0')
+		fn = av[0];
+	if (ac > 1 && *av[1] != '\0')
+		size = av[1];
+	if (ac > 2 && *av[2] != '\0') {
+
+		r = str2bytes(av[2], &page_size, 0);
+		if (r != NULL)
+			ARGV_ERR("(-sfile) granularity \"%s\": %s\n", av[2], r);
+	}
+
+	AN(fn);
+	AN(size);
 
 	sc = calloc(sizeof *sc, 1);
 	XXXAN(sc);
@@ -271,83 +274,60 @@ smf_init(struct stevedore *parent, const char *spec)
 	for (u = 0; u < NBUCKET; u++)
 		VTAILQ_INIT(&sc->free[u]);
 	VTAILQ_INIT(&sc->used);
-	sc->pagesize = getpagesize();
+	sc->pagesize = page_size;
 
 	parent->priv = sc;
 
-	/* If no size specified, use 50% of filesystem free space */
-	if (spec == NULL || *spec == '\0')
-		asprintf(&p, ".,50%%");
-	else if (strchr(spec, ',') == NULL)
-		asprintf(&p, "%s,", spec);
-	else
-		p = strdup(spec);
-	XXXAN(p);
-	size = strchr(p, ',');
-	XXXAN(size);
-
-	*size++ = '\0';
-
 	/* try to create a new file of this name */
-	sc->fd = open(p, O_RDWR | O_CREAT | O_EXCL, 0600);
+#ifdef O_LARGEFILE
+	sc->fd = open(fn, O_RDWR | O_CREAT | O_EXCL | O_LARGEFILE, 0600);
+#else
+	sc->fd = open(fn, O_RDWR | O_CREAT | O_EXCL, 0600);
+#endif
 	if (sc->fd >= 0) {
-		sc->filename = p;
+		sc->filename = fn;
+		mgt_child_inherit(sc->fd, "storage_file");
 		smf_initfile(sc, size, 1);
 		return;
 	}
 
 	/* it must exist then */
-	if (stat(p, &st)) {
-		fprintf(stderr,
-		    "Error: (-sfile) \"%s\" "
-		    "does not exist and could not be created\n", p);
-		exit (2);
-	}
+	if (stat(fn, &st))
+		ARGV_ERR("(-sfile) \"%s\" "
+		    "does not exist and could not be created\n", fn);
 
 	/* and it should be a file or directory */
-	if (!(S_ISREG(st.st_mode) || S_ISDIR(st.st_mode))) {
-		fprintf(stderr,
-		    "Error: (-sfile) \"%s\" "
-		    "is neither file nor directory\n", p);
-		exit (2);
-	}
+	if (!(S_ISREG(st.st_mode) || S_ISDIR(st.st_mode)))
+		ARGV_ERR("(-sfile) \"%s\" is neither file nor directory\n", fn);
 
 	if (S_ISREG(st.st_mode)) {
-		sc->fd = open(p, O_RDWR);
-		if (sc->fd < 0) {
-			fprintf(stderr,
-			    "Error: (-sfile) \"%s\" "
-			    "could not open (%s)\n", p, strerror(errno));
-			exit (2);
-		}
+		sc->fd = open(fn, O_RDWR);
+		if (sc->fd < 0)
+			ARGV_ERR("(-sfile) \"%s\" could not open (%s)\n",
+			    fn, strerror(errno));
 		AZ(fstat(sc->fd, &st));
-		if (!S_ISREG(st.st_mode)) {
-			fprintf(stderr,
-			    "Error: (-sfile) \"%s\" "
-			    "was not a file after opening\n", p);
-			exit (2);
-		}
-		sc->filename = p;
+		if (!S_ISREG(st.st_mode))
+			ARGV_ERR("(-sfile) \"%s\" "
+			    "was not a file after opening\n", fn);
+		sc->filename = fn;
+		mgt_child_inherit(sc->fd, "storage_file");
 		smf_initfile(sc, size, 0);
 		return;
 	}
 
-
-	asprintf(&q, "%s/varnish.XXXXXX", p);
+	asprintf(&q, "%s/varnish.XXXXXX", fn);
 	XXXAN(q);
 	sc->fd = mkstemp(q);
-	if (sc->fd < 0) {
-		fprintf(stderr,
-		    "Error: (-sfile) \"%s\" "
-		    "mkstemp(%s) failed (%s)\n", p, q, strerror(errno));
-		exit (2);
-	}
+	if (sc->fd < 0)
+		ARGV_ERR("(-sfile) \"%s\" "
+		    "mkstemp(%s) failed (%s)\n", fn, q, strerror(errno));
 	AZ(unlink(q));
-	asprintf(&sc->filename, "%s (unlinked)", q);
-	XXXAN(sc->filename);
+	asprintf(&p, "%s (unlinked)", q);
+	XXXAN(p);
+	sc->filename = p;
 	free(q);
 	smf_initfile(sc, size, 1);
-	free(p);
+	mgt_child_inherit(sc->fd, "storage_file");
 }
 
 /*--------------------------------------------------------------------
@@ -363,6 +343,7 @@ insfree(struct smf_sc *sc, struct smf *sp)
 
 	assert(sp->alloc == 0);
 	assert(sp->flist == NULL);
+	Lck_AssertHeld(&sc->mtx);
 	b = sp->size / sc->pagesize;
 	if (b >= NBUCKET) {
 		b = NBUCKET - 1;
@@ -386,12 +367,13 @@ insfree(struct smf_sc *sc, struct smf *sp)
 }
 
 static void
-remfree(struct smf_sc *sc, struct smf *sp)
+remfree(const struct smf_sc *sc, struct smf *sp)
 {
 	size_t b;
 
 	assert(sp->alloc == 0);
 	assert(sp->flist != NULL);
+	Lck_AssertHeld(&sc->mtx);
 	b = sp->size / sc->pagesize;
 	if (b >= NBUCKET) {
 		b = NBUCKET - 1;
@@ -614,7 +596,7 @@ smf_open_chunk(struct smf_sc *sc, off_t sz, off_t off, off_t *fail, off_t *sum)
 }
 
 static void
-smf_open(struct stevedore *st)
+smf_open(const struct stevedore *st)
 {
 	struct smf_sc *sc;
 	off_t fail = 1 << 30;	/* XXX: where is OFF_T_MAX ? */
@@ -622,14 +604,16 @@ smf_open(struct stevedore *st)
 
 	sc = st->priv;
 
+	Lck_New(&sc->mtx);
+	Lck_Lock(&sc->mtx);
 	smf_open_chunk(sc, sc->filesize, 0, &fail, &sum);
+	Lck_Unlock(&sc->mtx);
 	printf("managed to mmap %ju bytes of %ju\n",
 	    (uintmax_t)sum, sc->filesize);
 
 	/* XXX */
-	if (sum < MINPAGES * (uintmax_t)getpagesize())
+	if (sum < MINPAGES * (off_t)getpagesize())
 		exit (2);
-	MTX_INIT(&sc->mtx);
 
 	VSL_stats->sm_bfree += sc->filesize;
 }
@@ -645,18 +629,19 @@ smf_alloc(struct stevedore *st, size_t size)
 	assert(size > 0);
 	size += (sc->pagesize - 1);
 	size &= ~(sc->pagesize - 1);
-	LOCK(&sc->mtx);
+	Lck_Lock(&sc->mtx);
 	VSL_stats->sm_nreq++;
 	smf = alloc_smf(sc, size);
 	if (smf == NULL) {
-		UNLOCK(&sc->mtx);
+		Lck_Unlock(&sc->mtx);
 		return (NULL);
 	}
 	CHECK_OBJ_NOTNULL(smf, SMF_MAGIC);
 	VSL_stats->sm_nobj++;
 	VSL_stats->sm_balloc += smf->size;
 	VSL_stats->sm_bfree -= smf->size;
-	UNLOCK(&sc->mtx);
+	Lck_Unlock(&sc->mtx);
+	CHECK_OBJ_NOTNULL(&smf->s, STORAGE_MAGIC);	/*lint !e774 */
 	XXXAN(smf);
 	assert(smf->size == size);
 	smf->s.space = size;
@@ -666,14 +651,13 @@ smf_alloc(struct stevedore *st, size_t size)
 	smf->s.stevedore = st;
 	smf->s.fd = smf->sc->fd;
 	smf->s.where = smf->offset;
-	CHECK_OBJ_NOTNULL(&smf->s, STORAGE_MAGIC);
 	return (&smf->s);
 }
 
 /*--------------------------------------------------------------------*/
 
 static void
-smf_trim(struct storage *s, size_t size)
+smf_trim(const struct storage *s, size_t size)
 {
 	struct smf *smf;
 	struct smf_sc *sc;
@@ -688,18 +672,19 @@ smf_trim(struct storage *s, size_t size)
 	size += (sc->pagesize - 1);
 	size &= ~(sc->pagesize - 1);
 	if (smf->size > size) {
-		LOCK(&sc->mtx);
+		Lck_Lock(&sc->mtx);
 		VSL_stats->sm_balloc -= (smf->size - size);
 		VSL_stats->sm_bfree += (smf->size - size);
 		trim_smf(smf, size);
 		assert(smf->size == size);
-		UNLOCK(&sc->mtx);
+		Lck_Unlock(&sc->mtx);
 		smf->s.space = size;
 	}
 }
 
 /*--------------------------------------------------------------------*/
 
+/*lint -e{818} not const-able */
 static void
 smf_free(struct storage *s)
 {
@@ -709,23 +694,24 @@ smf_free(struct storage *s)
 	CHECK_OBJ_NOTNULL(s, STORAGE_MAGIC);
 	CAST_OBJ_NOTNULL(smf, s->priv, SMF_MAGIC);
 	sc = smf->sc;
-	LOCK(&sc->mtx);
+	Lck_Lock(&sc->mtx);
 	VSL_stats->sm_nobj--;
 	VSL_stats->sm_balloc -= smf->size;
 	VSL_stats->sm_bfree += smf->size;
 	free_smf(smf);
-	UNLOCK(&sc->mtx);
+	Lck_Unlock(&sc->mtx);
 }
 
 /*--------------------------------------------------------------------*/
 
 struct stevedore smf_stevedore = {
-	.name =		"file",
-	.init =		smf_init,
-	.open =		smf_open,
-	.alloc =	smf_alloc,
-	.trim =		smf_trim,
-	.free =		smf_free,
+	.magic	=	STEVEDORE_MAGIC,
+	.name	=	"file",
+	.init	=	smf_init,
+	.open	=	smf_open,
+	.alloc	=	smf_alloc,
+	.trim	=	smf_trim,
+	.free	=	smf_free,
 };
 
 #ifdef INCLUDE_TEST_DRIVER
