@@ -1,6 +1,6 @@
 /*-
  * Copyright (c) 2006 Verdens Gang AS
- * Copyright (c) 2006-2007 Linpro AS
+ * Copyright (c) 2006-2008 Linpro AS
  * All rights reserved.
  *
  * Author: Poul-Henning Kamp <phk@phk.freebsd.dk>
@@ -26,11 +26,13 @@
  * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
  * SUCH DAMAGE.
  *
- * $Id: varnishlog.c 2296 2007-12-18 16:13:36Z des $
+ * $Id: varnishlog.c 3625 2009-02-05 12:36:41Z tfheen $
  *
  * Log tailer for Varnish
  */
 
+#include "config.h"
+
 #include <errno.h>
 #include <fcntl.h>
 #include <regex.h>
@@ -45,12 +47,6 @@
 #include "compat/daemon.h"
 #endif
 
-#ifdef HAVE_VIS_H
-#include <vis.h>
-#else
-#include "compat/vis.h"
-#endif
-
 #include "vsb.h"
 #include "vpf.h"
 
@@ -80,12 +76,25 @@ name2tag(const char *n)
 
 static struct vsb	*ob[65536];
 static unsigned char	flg[65536];
+static enum shmlogtag   last[65536];
 #define F_INVCL		(1 << 0)
 #define F_MATCH		(1 << 1)
 
 static int		match_tag = -1;
 static regex_t		match_re;
 
+static void
+h_order_finish(int fd)
+{
+
+	vsb_finish(ob[fd]);
+	if (vsb_len(ob[fd]) > 1 &&
+	    (match_tag == -1 || flg[fd] & F_MATCH))
+		printf("%s\n", vsb_data(ob[fd]));
+	flg[fd] &= ~F_MATCH;
+	vsb_clear(ob[fd]);
+}
+
 static void
 clean_order(void)
 {
@@ -104,23 +113,49 @@ clean_order(void)
 }
 
 static int
-h_order(void *priv, enum shmlogtag tag, unsigned fd, unsigned len, unsigned spec, const char *ptr)
+h_order(void *priv, enum shmlogtag tag, unsigned fd, unsigned len,
+    unsigned spec, const char *ptr)
 {
+	char type;
 
 	(void)priv;
 
+	type = (spec & VSL_S_CLIENT) ? 'c' :
+	    (spec & VSL_S_BACKEND) ? 'b' : '-';
+
 	if (!(spec & (VSL_S_CLIENT|VSL_S_BACKEND))) {
 		if (!b_flag && !c_flag)
 			VSL_H_Print(stdout, tag, fd, len, spec, ptr);
 		return (0);
 	}
 	if (ob[fd] == NULL) {
-		ob[fd] = vsb_new(NULL, NULL, 0, VSB_AUTOEXTEND);
+		ob[fd] = vsb_newauto();
 		assert(ob[fd] != NULL);
 	}
 	if (tag == match_tag &&
 	    !regexec(&match_re, ptr, 0, NULL, 0))
 		flg[fd] |= F_MATCH;
+
+	if ((tag == SLT_BackendOpen || tag == SLT_SessionOpen ||
+		(tag == SLT_ReqStart &&
+		    last[fd] != SLT_SessionOpen &&
+		    last[fd] != SLT_VCL_acl) ||
+		(tag == SLT_BackendXID &&
+		    last[fd] != SLT_BackendOpen)) &&
+	    vsb_len(ob[fd]) != 0) {
+		/*
+		 * This is the start of a new request, yet we haven't seen
+		 * the end of the previous one.  Spit it out anyway before
+		 * starting on the new one.
+		 */
+		if (last[fd] != SLT_SessionClose)
+			vsb_printf(ob[fd], "%5d %-12s %c %s\n",
+			    fd, "Interrupted", type, VSL_tags[tag]);
+		h_order_finish(fd);
+	}
+
+	last[fd] = tag;
+
 	switch (tag) {
 	case SLT_VCL_call:
 		if (flg[fd] & F_INVCL)
@@ -128,10 +163,7 @@ h_order(void *priv, enum shmlogtag tag, unsigned fd, unsigned len, unsigned spec
 		else
 			flg[fd] |= F_INVCL;
 		vsb_printf(ob[fd], "%5d %-12s %c %.*s",
-		    fd, VSL_tags[tag],
-		    ((spec & VSL_S_CLIENT) ? 'c' : \
-		    (spec & VSL_S_BACKEND) ? 'b' : '-'),
-		    len, ptr);
+		    fd, VSL_tags[tag], type, len, ptr);
 		return (0);
 	case SLT_VCL_trace:
 	case SLT_VCL_return:
@@ -149,20 +181,13 @@ h_order(void *priv, enum shmlogtag tag, unsigned fd, unsigned len, unsigned spec
 		flg[fd] &= ~F_INVCL;
 	}
 	vsb_printf(ob[fd], "%5d %-12s %c %.*s\n",
-	    fd, VSL_tags[tag],
-	    ((spec & VSL_S_CLIENT) ? 'c' : (spec & VSL_S_BACKEND) ? 'b' : '-'),
-	    len, ptr);
+	    fd, VSL_tags[tag], type, len, ptr);
 	switch (tag) {
 	case SLT_ReqEnd:
 	case SLT_BackendClose:
 	case SLT_BackendReuse:
 	case SLT_StatSess:
-		vsb_finish(ob[fd]);
-		if (vsb_len(ob[fd]) > 1 &&
-		    (match_tag == -1 || flg[fd] & F_MATCH))
-			printf("%s\n", vsb_data(ob[fd]));
-		flg[fd] &= ~F_MATCH;
-		vsb_clear(ob[fd]);
+		h_order_finish(fd);
 		break;
 	default:
 		break;
@@ -229,6 +254,9 @@ open_log(const char *w_arg, int a_flag)
 	int fd, flags;
 
 	flags = (a_flag ? O_APPEND : O_TRUNC) | O_WRONLY | O_CREAT;
+#ifdef O_LARGEFILE
+	flags |= O_LARGEFILE;
+#endif
 	if (!strcmp(w_arg, "-"))
 		fd = STDOUT_FILENO;
 	else
@@ -243,7 +271,7 @@ open_log(const char *w_arg, int a_flag)
 static void
 do_write(struct VSL_data *vd, const char *w_arg, int a_flag)
 {
-	int fd, i;
+	int fd, i, l;
 	unsigned char *p;
 
 	fd = open_log(w_arg, a_flag);
@@ -253,7 +281,8 @@ do_write(struct VSL_data *vd, const char *w_arg, int a_flag)
 		if (i < 0)
 			break;
 		if (i > 0) {
-			i = write(fd, p, 5 + p[1]);
+			l = SHMLOG_LEN(p);
+			i = write(fd, p, SHMLOG_NEXTTAG + l);
 			if (i < 0) {
 				perror(w_arg);
 				exit(1);
@@ -273,8 +302,8 @@ do_write(struct VSL_data *vd, const char *w_arg, int a_flag)
 static void
 usage(void)
 {
-	fprintf(stderr,
-	    "usage: varnishlog %s [-aDoV] [-n varnish_name] [-P file] [-w file]\n", VSL_USAGE);
+	fprintf(stderr, "usage: varnishlog "
+	    "%s [-aDoV] [-n varnish_name] [-P file] [-w file]\n", VSL_USAGE);
 	exit(1);
 }
 
@@ -282,7 +311,7 @@ int
 main(int argc, char **argv)
 {
 	int c;
-	int a_flag = 0, D_flag = 0, o_flag = 0;
+	int a_flag = 0, D_flag = 0, o_flag = 0, u_flag = 0;
 	const char *n_arg = NULL;
 	const char *P_arg = NULL;
 	const char *w_arg = NULL;
@@ -291,11 +320,19 @@ main(int argc, char **argv)
 
 	vd = VSL_New();
 
-	while ((c = getopt(argc, argv, VSL_ARGS "aDon:P:Vw:")) != -1) {
+	while ((c = getopt(argc, argv, VSL_ARGS "aDon:P:uVw:")) != -1) {
 		switch (c) {
 		case 'a':
 			a_flag = 1;
 			break;
+		case 'b':
+			b_flag = 1;
+			VSL_Arg(vd, c, optarg);
+			break;
+		case 'c':
+			c_flag = 1;
+			VSL_Arg(vd, c, optarg);
+			break;
 		case 'D':
 			D_flag = 1;
 			break;
@@ -308,22 +345,15 @@ main(int argc, char **argv)
 		case 'P':
 			P_arg = optarg;
 			break;
+		case 'u':
+			u_flag = 1;
+			break;
 		case 'V':
 			varnish_version("varnishlog");
 			exit(0);
 		case 'w':
 			w_arg = optarg;
 			break;
-		case 'c':
-			c_flag = 1;
-			if (VSL_Arg(vd, c, optarg) > 0)
-				break;
-			usage();
-		case 'b':
-			b_flag = 1;
-			if (VSL_Arg(vd, c, optarg) > 0)
-				break;
-			usage();
 		default:
 			if (VSL_Arg(vd, c, optarg) > 0)
 				break;
@@ -337,7 +367,7 @@ main(int argc, char **argv)
 	if (VSL_OpenLog(vd, n_arg))
 		exit(1);
 
-	if (P_arg && (pfh = vpf_open(P_arg, 0600, NULL)) == NULL) {
+	if (P_arg && (pfh = vpf_open(P_arg, 0644, NULL)) == NULL) {
 		perror(P_arg);
 		exit(1);
 	}
@@ -358,6 +388,9 @@ main(int argc, char **argv)
 	if (o_flag)
 		do_order(vd, argc - optind, argv + optind);
 
+	if (u_flag)
+		setbuf(stdout, NULL);
+
 	while (VSL_Dispatch(vd, VSL_H_Print, stdout) >= 0) {
 		if (fflush(stdout) != 0) {
 			perror("stdout");
