@@ -1,6 +1,6 @@
 /*-
  * Copyright (c) 2006 Verdens Gang AS
- * Copyright (c) 2006-2007 Linpro AS
+ * Copyright (c) 2006-2008 Linpro AS
  * All rights reserved.
  *
  * Author: Anders Berg <andersb@vgnett.no>
@@ -27,7 +27,7 @@
  * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
  * SUCH DAMAGE.
  *
- * $Id: varnishncsa.c 2325 2007-12-20 13:57:44Z des $
+ * $Id: varnishncsa.c 3586 2009-02-05 09:17:38Z tfheen $
  *
  * Obtain log data from the shared memory log, order it by session ID, and
  * display it in Apache / NCSA combined log format:
@@ -59,6 +59,8 @@
  *		- Maybe rotate/compress log
  */
 
+#include "config.h"
+
 #include <ctype.h>
 #include <errno.h>
 #include <signal.h>
@@ -88,20 +90,22 @@ static struct logline {
 	char *df_Referer;		/* %{Referer}i */
 	char *df_Uq;			/* %U%q, URL path and query string */
 	char *df_User_agent;		/* %{User-agent}i */
+	char *df_X_Forwarded_For;	/* %{X-Forwarded-For}i */
 	char *df_b;			/* %b, Bytes */
 	char *df_h;			/* %h (host name / IP adress)*/
 	char *df_m;			/* %m, Request method*/
 	char *df_s;			/* %s, Status */
+	struct tm df_t;			/* %t, Date and time */
 	char *df_u;			/* %u, Remote user */
 	int bogus;			/* bogus request */
 } **ll;
 
 static size_t nll;
-
-static time_t t;
+static int prefer_x_forwarded_for = 0;
 
 static int
-isprefix(const char *str, const char *prefix, const char *end, const char **next)
+isprefix(const char *str, const char *prefix, const char *end,
+    const char **next)
 {
 
 	while (str < end && *str && *prefix &&
@@ -174,22 +178,188 @@ trimline(const char *str, const char *end)
 	return (p);
 }
 
+static int
+collect_backend(struct logline *lp, enum shmlogtag tag, unsigned spec,
+    const char *ptr, unsigned len)
+{
+	const char *end, *next;
+
+	assert(spec & VSL_S_BACKEND);
+	end = ptr + len;
+
+	switch (tag) {
+	case SLT_BackendOpen:
+		if (lp->df_h != NULL)
+			lp->bogus = 1;
+		else
+			if (isprefix(ptr, "default", end, &next))
+				lp->df_h = trimfield(next, end);
+			else
+				lp->df_h = trimfield(ptr, end);
+		break;
+
+	case SLT_TxRequest:
+		if (lp->df_m != NULL)
+			lp->bogus = 1;
+		else
+			lp->df_m = trimline(ptr, end);
+		break;
+
+	case SLT_TxURL:
+		if (lp->df_Uq != NULL)
+			lp->bogus = 1;
+		else
+			lp->df_Uq = trimline(ptr, end);
+		break;
+
+	case SLT_TxProtocol:
+		if (lp->df_H != NULL)
+			lp->bogus = 1;
+		else
+			lp->df_H = trimline(ptr, end);
+		break;
+
+	case SLT_RxStatus:
+		if (lp->df_s != NULL)
+			lp->bogus = 1;
+		else
+			lp->df_s = trimline(ptr, end);
+		break;
+
+	case SLT_RxHeader:
+		if (isprefix(ptr, "content-length:", end, &next))
+			lp->df_b = trimline(next, end);
+		else if (isprefix(ptr, "date:", end, &next) &&
+		    strptime(next, "%a, %d %b %Y %T", &lp->df_t) == NULL)
+			lp->bogus = 1;
+		break;
+
+	case SLT_TxHeader:
+		if (isprefix(ptr, "user-agent:", end, &next))
+			lp->df_User_agent = trimline(next, end);
+		else if (isprefix(ptr, "referer:", end, &next))
+			lp->df_Referer = trimline(next, end);
+		else if (isprefix(ptr, "authorization:", end, &next) &&
+		    isprefix(next, "basic", end, &next))
+			lp->df_u = trimline(next, end);
+		else if (isprefix(ptr, "x-forwarded-for:", end, &next))
+			lp->df_X_Forwarded_For = trimline(next, end);
+		else if (isprefix(ptr, "host:", end, &next))
+			lp->df_Host = trimline(next, end);
+		break;
+
+	case SLT_BackendReuse:
+	case SLT_BackendClose:
+		/* got it all */
+		return (0);
+
+	default:
+		break;
+	}
+
+	/* more to come */
+	return (1);
+}
+
+static int
+collect_client(struct logline *lp, enum shmlogtag tag, unsigned spec,
+    const char *ptr, unsigned len)
+{
+	const char *end, *next;
+	long l;
+	time_t t;
+
+	assert(spec & VSL_S_CLIENT);
+	end = ptr + len;
+
+	switch (tag) {
+	case SLT_ReqStart:
+		if (lp->df_h != NULL)
+			lp->bogus = 1;
+		else
+			lp->df_h = trimfield(ptr, end);
+		break;
+
+	case SLT_RxRequest:
+		if (lp->df_m != NULL)
+			lp->bogus = 1;
+		else
+			lp->df_m = trimline(ptr, end);
+		break;
+
+	case SLT_RxURL:
+		if (lp->df_Uq != NULL)
+			lp->bogus = 1;
+		else
+			lp->df_Uq = trimline(ptr, end);
+		break;
+
+	case SLT_RxProtocol:
+		if (lp->df_H != NULL)
+			lp->bogus = 1;
+		else
+			lp->df_H = trimline(ptr, end);
+		break;
+
+	case SLT_TxStatus:
+		if (lp->df_s != NULL)
+			lp->bogus = 1;
+		else
+			lp->df_s = trimline(ptr, end);
+		break;
+
+	case SLT_RxHeader:
+		if (isprefix(ptr, "user-agent:", end, &next))
+			lp->df_User_agent = trimline(next, end);
+		else if (isprefix(ptr, "referer:", end, &next))
+			lp->df_Referer = trimline(next, end);
+		else if (isprefix(ptr, "authorization:", end, &next) &&
+		    isprefix(next, "basic", end, &next))
+			lp->df_u = trimline(next, end);
+		else if (isprefix(ptr, "x-forwarded-for:", end, &next))
+			lp->df_X_Forwarded_For = trimline(next, end);
+		else if (isprefix(ptr, "host:", end, &next))
+			lp->df_Host = trimline(next, end);
+		break;
+
+	case SLT_Length:
+		if (lp->df_b != NULL)
+			lp->bogus = 1;
+		else
+			lp->df_b = trimline(ptr, end);
+		break;
+
+	case SLT_SessionClose:
+		if (strncmp(ptr, "pipe", len) == 0 ||
+		    strncmp(ptr, "error", len) == 0)
+			lp->bogus = 1;
+		break;
+
+	case SLT_ReqEnd:
+		if (sscanf(ptr, "%*u %*u.%*u %ld.", &l) != 1) {
+			lp->bogus = 1;
+		} else {
+			t = l;
+			localtime_r(&t, &lp->df_t);
+		}
+		/* got it all */
+		return (0);
+
+	default:
+		break;
+	}
+
+	/* more to come */
+	return (1);
+}
+
 static int
 h_ncsa(void *priv, enum shmlogtag tag, unsigned fd,
     unsigned len, unsigned spec, const char *ptr)
 {
-	const char *end, *next;
-	char *q;
-	FILE *fo;
-	long l;
-	struct tm tm;
-	char tbuf[40];
 	struct logline *lp;
-
-	end = ptr + len;
-
-	if (!(spec & VSL_S_CLIENT || spec & VSL_S_BACKEND))
-		return (reopen);
+	FILE *fo = priv;
+	char *q, tbuf[64];
 
 	if (fd >= nll) {
 		struct logline **newll = ll;
@@ -209,126 +379,22 @@ h_ncsa(void *priv, enum shmlogtag tag, unsigned fd,
 	}
 	lp = ll[fd];
 
-	switch (tag) {
-	case SLT_BackendOpen:
-		if (!(spec & VSL_S_BACKEND))
-			break;
-		if (lp->df_h != NULL)
-			lp->bogus = 1;
-		else {
-			if (isprefix(ptr, "default", end, &next))
-				lp->df_h = trimfield(next, end);
-			else
-				lp->df_h = trimfield(ptr, end);
-		}
-		break;
-	case SLT_ReqStart:
-		if (lp->df_h != NULL)
-			lp->bogus = 1;
-		else
-			lp->df_h = trimfield(ptr, end);
-		break;
-
-	case SLT_TxRequest:
-		if (!(spec & VSL_S_BACKEND))
-			break;
-	case SLT_RxRequest:
-		if (tag == SLT_RxRequest && (spec & VSL_S_BACKEND))
-			break;
-
-		if (lp->df_m != NULL)
-			lp->bogus = 1;
-		else
-			lp->df_m = trimline(ptr, end);
-		break;
-
-	case SLT_TxURL:
-		if (!(spec & VSL_S_BACKEND))
-			break;
-	case SLT_RxURL:
-		if (tag == SLT_RxURL && (spec & VSL_S_BACKEND))
-			break;
-
-		if (lp->df_Uq != NULL)
-			lp->bogus = 1;
-		else
-			lp->df_Uq = trimline(ptr, end);
-		break;
-
-	case SLT_TxProtocol:
-		if (!(spec & VSL_S_BACKEND))
-			break;
-	case SLT_RxProtocol:
-		if (tag == SLT_RxProtocol && (spec & VSL_S_BACKEND))
-			break;
-
-		if (lp->df_H != NULL)
-			lp->bogus = 1;
-		else
-			lp->df_H = trimline(ptr, end);
-		break;
-
-	case SLT_RxStatus:
-		if (!(spec & VSL_S_BACKEND))
-			break;
-	case SLT_TxStatus:
-		if (tag == SLT_TxStatus && (spec & VSL_S_BACKEND))
-			break;
-
-		if (lp->df_s != NULL)
-			lp->bogus = 1;
-		else
-			lp->df_s = trimline(ptr, end);
-		break;
-
-	case SLT_TxHeader:
-		if (!(spec & VSL_S_BACKEND))
-			break;
-	case SLT_RxHeader:
-		if (tag == SLT_RxHeader && (spec & VSL_S_BACKEND)) {
-			if (isprefix(ptr, "content-length:", end, &next)) {
-				lp->df_b = trimline(next, end);
-			} else if (isprefix(ptr, "date:", end, &next)) {
-				if (strptime(trimline(next, end), "%a, %d %b %Y %T", &tm))
-					t = mktime(&tm);
-			}
-			break;
-		}
-		if (isprefix(ptr, "user-agent:", end, &next))
-			lp->df_User_agent = trimline(next, end);
-		else if (isprefix(ptr, "referer:", end, &next))
-			lp->df_Referer = trimline(next, end);
-		else if (isprefix(ptr, "authorization:", end, &next) &&
-		    isprefix(next, "basic", end, &next))
-			lp->df_u = trimline(next, end);
-		else if (isprefix(ptr, "host:", end, &next))
-			lp->df_Host = trimline(next, end);
-		break;
-
-	case SLT_Length:
-		if (lp->df_b != NULL)
-			lp->bogus = 1;
-		else
-			lp->df_b = trimline(ptr, end);
-		break;
-
-	default:
-		break;
+	if (spec & VSL_S_BACKEND) {
+		if (collect_backend(lp, tag, spec, ptr, len))
+			return (reopen);
+	} else if (spec & VSL_S_CLIENT) {
+		if (collect_client(lp, tag, spec, ptr, len))
+			return (reopen);
+	} else {
+		/* huh? */
+		return (reopen);
 	}
 
-	if ((spec & VSL_S_CLIENT) && tag != SLT_ReqEnd)
-		return (reopen);
-
-	if ((spec & VSL_S_BACKEND) && tag != SLT_BackendReuse &&
-	    (tag != SLT_BackendClose || lp->df_Uq))
-		return (reopen);
-
-	if (tag == SLT_ReqEnd) {
-		if (sscanf(ptr, "%*u %*u.%*u %ld.", &l) != 1)
-			lp->bogus = 1;
-		else
-			t = l;
-	}
+#if 0
+	/* non-optional fields */
+	if (!lp->df_m || !lp->df_Uq || !lp->df_H || !lp->df_s)
+		lp->bogus = 1;
+#endif
 
 	if (!lp->bogus) {
 		fo = priv;
@@ -336,6 +402,8 @@ h_ncsa(void *priv, enum shmlogtag tag, unsigned fd,
 		/* %h */
 		if (!lp->df_h && spec & VSL_S_BACKEND)
 			fprintf(fo, "127.0.0.1 ");
+		else if (lp->df_X_Forwarded_For && prefer_x_forwarded_for)
+			fprintf(fo, "%s ", lp->df_X_Forwarded_For);
 		else
 			fprintf(fo, "%s ", lp->df_h ? lp->df_h : "-");
 
@@ -362,8 +430,7 @@ h_ncsa(void *priv, enum shmlogtag tag, unsigned fd,
 		}
 
 		/* %t */
-		localtime_r(&t, &tm);
-		strftime(tbuf, sizeof tbuf, "[%d/%b/%Y:%T %z]", &tm);
+		strftime(tbuf, sizeof tbuf, "[%d/%b/%Y:%T %z]", &lp->df_t);
 		fprintf(fo, "%s ", tbuf);
 
 		/*
@@ -374,7 +441,7 @@ h_ncsa(void *priv, enum shmlogtag tag, unsigned fd,
 		if (lp->df_Host) {
 			if (strncmp(lp->df_Host, "http://", 7) != 0)
 				fprintf(fo, "http://");
-			fprintf(fo, lp->df_Host);
+			fprintf(fo, "%s", lp->df_Host);
 		}
 		fprintf(fo, "%s ", lp->df_Uq);
 		fprintf(fo, "%s\" ", lp->df_H);
@@ -383,7 +450,7 @@ h_ncsa(void *priv, enum shmlogtag tag, unsigned fd,
 		fprintf(fo, "%s ", lp->df_s);
 
 		/* %b */
-		fprintf(fo, "%s ", lp->df_b);
+		fprintf(fo, "%s ", lp->df_b ? lp->df_b : "-");
 
 		/* %{Referer}i */
 		fprintf(fo, "\"%s\" ",
@@ -393,9 +460,8 @@ h_ncsa(void *priv, enum shmlogtag tag, unsigned fd,
 		fprintf(fo, "\"%s\"\n",
 		    lp->df_User_agent ? lp->df_User_agent : "-");
 
-		/* hack: flush after every line if writing to file */
-		if (fo != stdout)
-			fflush(fo);
+		/* flush the stream */
+		fflush(fo);
 	}
 
 	/* clean up */
@@ -405,13 +471,14 @@ h_ncsa(void *priv, enum shmlogtag tag, unsigned fd,
 	freez(lp->df_Referer);
 	freez(lp->df_Uq);
 	freez(lp->df_User_agent);
+	freez(lp->df_X_Forwarded_For);
 	freez(lp->df_b);
 	freez(lp->df_h);
 	freez(lp->df_m);
 	freez(lp->df_s);
 	freez(lp->df_u);
 #undef freez
-	lp->bogus = 0;
+	memset(lp, 0, sizeof *lp);
 
 	return (reopen);
 }
@@ -444,7 +511,9 @@ static void
 usage(void)
 {
 
-	fprintf(stderr, "usage: varnishncsa %s [-aDV] [-n varnish_name] [-P file] [-w file]\n", VSL_ARGS);
+	fprintf(stderr,
+	    "usage: varnishncsa %s [-aDV] [-n varnish_name] "
+	    "[-P file] [-w file]\n", VSL_USAGE);
 	exit(1);
 }
 
@@ -462,11 +531,14 @@ main(int argc, char *argv[])
 
 	vd = VSL_New();
 
-	while ((c = getopt(argc, argv, VSL_ARGS "aDn:P:Vw:")) != -1) {
+	while ((c = getopt(argc, argv, VSL_ARGS "aDn:P:Vw:f")) != -1) {
 		switch (c) {
 		case 'a':
 			a_flag = 1;
 			break;
+		case 'f':
+			prefer_x_forwarded_for = 1;
+			break;
 		case 'D':
 			D_flag = 1;
 			break;
@@ -492,7 +564,7 @@ main(int argc, char *argv[])
 	if (VSL_OpenLog(vd, n_arg))
 		exit(1);
 
-	if (P_arg && (pfh = vpf_open(P_arg, 0600, NULL)) == NULL) {
+	if (P_arg && (pfh = vpf_open(P_arg, 0644, NULL)) == NULL) {
 		perror(P_arg);
 		exit(1);
 	}
