@@ -1,6 +1,6 @@
 /*-
  * Copyright (c) 2006 Verdens Gang AS
- * Copyright (c) 2006-2007 Linpro AS
+ * Copyright (c) 2006-2008 Linpro AS
  * All rights reserved.
  *
  * Author: Poul-Henning Kamp <phk@phk.freebsd.dk>
@@ -26,7 +26,7 @@
  * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
  * SUCH DAMAGE.
  *
- * $Id: cli_priv.h 1503 2007-06-07 09:40:20Z des $
+ * $Id: cli_priv.h 2982 2008-07-22 07:37:32Z phk $
  *
  * Varnish process internal CLI stuff.
  *
@@ -37,7 +37,7 @@
 
 struct cli;	/* NB: struct cli is opaque at this level.  */
 
-typedef void cli_func_t(struct cli*, char **av, void *priv);
+typedef void cli_func_t(struct cli*, const char * const *av, void *priv);
 
 struct cli_proto {
 	/* These must match the CLI_* macros in cli.h */
@@ -54,9 +54,12 @@ struct cli_proto {
 
 /* The implementation must provide these functions */
 void cli_out(struct cli *cli, const char *fmt, ...);
+void cli_quote(struct cli *cli, const char *str);
 void cli_param(struct cli *cli);
 void cli_result(struct cli *cli, unsigned r);
 
 /* From libvarnish/cli.c */
 void cli_dispatch(struct cli *cli, struct cli_proto *clp, const char *line);
 cli_func_t	cli_func_help;
+cli_func_t	cli_func_ping;
+struct cli_proto *cli_concat(struct cli_proto *, struct cli_proto *);
