@@ -1,6 +1,6 @@
 /*-
  * Copyright (c) 2006 Verdens Gang AS
- * Copyright (c) 2006-2007 Linpro AS
+ * Copyright (c) 2006-2008 Linpro AS
  * All rights reserved.
  *
  * Author: Poul-Henning Kamp <phk@phk.freebsd.dk>
@@ -26,7 +26,7 @@
  * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
  * SUCH DAMAGE.
  *
- * $Id: shmlog_tags.h 1693 2007-07-13 14:54:51Z des $
+ * $Id: shmlog_tags.h 3100 2008-08-18 09:10:11Z phk $
  *
  * Define the tags in the shared memory in a reusable format.
  * Whoever includes this get to define what the SLTM macro does.
@@ -84,6 +84,8 @@ SLTM(VCL_acl)
 SLTM(VCL_call)
 SLTM(VCL_trace)
 SLTM(VCL_return)
+SLTM(VCL_error)
+SLTM(VCL_info)
 SLTM(ReqStart)
 SLTM(Hit)
 SLTM(HitPass)
@@ -92,3 +94,10 @@ SLTM(ExpPick)
 SLTM(ExpKill)
 SLTM(WorkThread)
 SLTM(Terminate)
+
+SLTM(ESItrace)
+SLTM(ESI_xmlerror)
+
+SLTM(Hash)
+
+SLTM(Backend_health)
