@@ -1,6 +1,6 @@
 /*-
  * Copyright (c) 2006 Verdens Gang AS
- * Copyright (c) 2006-2007 Linpro AS
+ * Copyright (c) 2006-2008 Linpro AS
  * All rights reserved.
  *
  * Author: Poul-Henning Kamp <phk@phk.freebsd.dk>
@@ -26,7 +26,7 @@
  * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
  * SUCH DAMAGE.
  *
- * $Id: vcc_priv.h 1503 2007-06-07 09:40:20Z des $
+ * $Id: vcc_priv.h 3599 2009-02-05 10:13:52Z tfheen $
  *
  * Stuff shared between main.c and fixed_token.c
  */
@@ -39,8 +39,7 @@ struct vsb;
 #define isident(c) (isalpha(c) || isdigit(c) || (c) == '_' || (c) == '-')
 #define isvar(c) (isident(c) || (c) == '.')
 unsigned vcl_fixed_token(const char *p, const char **q);
-extern const char *vcl_tnames[256];
-void vcl_init_tnames(void);
+extern const char * const vcl_tnames[256];
 void vcl_output_lang_h(struct vsb *sb);
 
 #define PF(t)	(int)((t)->e - (t)->b), (t)->b
